#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# shpkpr documentation build configuration file, created by
# sphinx-quickstart on Tue Jul  9 22:26:36 2013.

import sys
import os

# Get the project root dir, which is the parent dir of this
cwd = os.getcwd()
project_root = os.path.dirname(cwd)

# Insert the project root dir as the first element in the PYTHONPATH.
# This lets us ensure that the source package is imported, and that its
# version is used.
sys.path.insert(0, project_root)

import shpkpr  # NOQA


# -- General configuration ---------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx_click.ext',
    'sphinxcontrib.programoutput',
]

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'shpkpr'
copyright = u'2017, ShopKeep.com Inc.'

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = shpkpr.__version__
# The full version, including alpha/beta/rc tags.
release = shpkpr.__version__

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'



# third-party imports
import pytest

# local imports
from shpkpr import exceptions


@exceptions.rewrap(ValueError)
def _this_doesnt_raise():
    """Dummy func for testing the exceptions.rewrap decorator
    """
    return 'something'


@exceptions.rewrap(ValueError)
def _this_raises_a_key_error():
    """Dummy func for testing the exceptions.rewrap decorator
    """
    return {}['non-existant-key']


@exceptions.rewrap(ValueError)
def _this_raises_a_value_error():
    """Dummy func for testing the exceptions.rewrap decorator
    """
    return int("banana for scale")


@exceptions.rewrap(ValueError, KeyError)
def _this_also_raises_a_value_error():
    """Dummy func for testing the exceptions.rewrap decorator
    """
    return int("I suggest you drop it, Mr. Data")


class ExceptionRaiser(object):
    """Dummy class for testing the exceptions.rewrap decorator
    """

    @exceptions.rewrap(KeyError)
    def this_raises_a_key_error(self):
        return {}['non-existant-key']

    @exceptions.rewrap(ValueError)
    def this_also_raises_a_key_error(self):
        return {}['non-existant-key']


def test_rewrap_func_doesnt_rewrap_uncaught_exception():
    with pytest.raises(KeyError):
        _this_raises_a_key_error()


def test_rewrap_func_raises_correct_exception():
    with pytest.raises(exceptions.ShpkprException):
        _this_raises_a_value_error()


def test_rewrap_func_raises_correct_exception_keyerror():
    with pytest.raises(KeyError):
        _this_also_raises_a_value_error()


def test_rewrap_method_raises_correct_exception():
    obj = ExceptionRaiser()
    with pytest.raises(exceptions.ShpkprException):
        obj.this_raises_a_key_error()


def test_rewrap_method_doesnt_rewrap_uncaught_exception():
    obj = ExceptionRaiser()
    with pytest.raises(KeyError):
        obj.this_also_raises_a_key_error()


def test_decorated_func_returns_as_normal_when_nothing_is_raised():
    assert _this_doesnt_raise() == 'something'



# stdlib imports
import json
import os

# third-party imports
import pytest


@pytest.fixture
def file_fixture():
    def _file_fixture(name):
        here = os.path.dirname(os.path.realpath(__file__))
        path = os.path.join(here, "fixtures", name)
        with open(path, 'r') as f:
            fixture = f.read()
        return fixture
    return _file_fixture


@pytest.fixture
def json_fixture(file_fixture):
    def _json_fixture(name):
        return json.loads(file_fixture(name + ".json"))
    return _json_fixture



import mock

from shpkpr.vault import resolve_secrets


@mock.patch("shpkpr.cli.options.hvac.Client")
def test_resolve_secrets(mock_vault_client_class):
    mock_vault_data = {
        'secret/my_project/my_path': {
            'my_key': 'some_secret_info'
        }
    }
    mock_rendered_template = {
        'secrets': {
            'MY_SECRET_USING_REL_PATH': {'source': 'my_project/my_path:my_key'},
            'MY_SECRET_USING_FULL_PATH': {'source': 'secret/my_project/my_path:my_key'},
        }
    }

    def read_vault_data(path):
        secrets = mock_vault_data.get(path, None)
        return dict(data=secrets) if secrets else None

    mock_vault_client = mock_vault_client_class.return_value
    mock_vault_client.read.side_effect = read_vault_data

    result = resolve_secrets(mock_vault_client, mock_rendered_template)
    assert 'MY_SECRET_USING_REL_PATH' not in result
    assert result['MY_SECRET_USING_FULL_PATH'] == 'some_secret_info'



# stdlib imports
import logging
import sys


def configure():
    """Configure logging for CLI use.

    This configures logging to go stdout by default and to only include the raw
    message in the output (no decoration).

    This configuration is CLI specific and other users of this library may
    choose to configure logging in a different way that's more appropriate for
    the use case (e.g. JSON output for server-side logging).
    """
    FORMAT_STRING = "%(message)s"
    LEVEL = logging.INFO
    STREAM = sys.stdout

    root_logger = logging.getLogger()
    handler = logging.StreamHandler(stream=STREAM)

    handler.setFormatter(logging.Formatter(FORMAT_STRING))
    root_logger.addHandler(handler)
    root_logger.setLevel(LEVEL)



# third-party imports
import mock
import responses


def test_no_args(runner):
    result = runner(['apps', 'deploy'])

    assert result.exit_code == 2
    assert 'Usage:' in result.output


def test_help(runner):
    result = runner(['apps', 'deploy', '--help'])

    assert result.exit_code == 0
    assert 'Usage:' in result.output
    assert 'Deploy one or more applications to Marathon.' in result.output


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.wait')
def test_no_force(mock_deployment_wait, runner, json_fixture):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=201,
                  json=json_fixture("deployment"),
                  match_querystring=True)
    mock_deployment_wait.return_value = True

    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_DOCKER_EXPOSED_PORT': '8080',
        'SHPKPR_DEPLOY_DOMAIN': 'mydomain.com',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    result = runner(['apps', 'deploy', '--template', _tmpl_path, 'RANDOM_LABEL=some_value'], env=env)

    assert result.exit_code == 0


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.wait')
def test_force(mock_deployment_wait, runner, json_fixture):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app?force=true',
                  status=201,
                  json=json_fixture("deployment"),
                  match_querystring=True)
    mock_deployment_wait.return_value = True

    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_DOCKER_EXPOSED_PORT': '8080',
        'SHPKPR_DEPLOY_DOMAIN': 'mydomain.com',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    result = runner(['apps', 'deploy', '--template', _tmpl_path, '--force', 'RANDOM_LABEL=some_value'], env=env)

    assert result.exit_code == 0


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.wait')
def test_multiple_templates(mock_deployment_wait, runner, json_fixture):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/',
                  status=201,
                  json=json_fixture("deployment"),
                  match_querystring=True)
    mock_deployment_wait.return_value = True

    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_APPLICATION_2': 'test-app-2',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_DOCKER_EXPOSED_PORT': '8080',
        'SHPKPR_DEPLOY_DOMAIN': 'mydomain.com',
        'SHPKPR_RANDOM_LABEL': 'some_value',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    _tmpl_path_2 = "tests/fixtures/templates/marathon/test-2.json.tmpl"
    result = runner(['apps', 'deploy', '--template', _tmpl_path, '--template', _tmpl_path_2], env=env)

    assert result.exit_code == 0


@responses.activate
def test_dry_run(runner, json_fixture):
    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_DOCKER_EXPOSED_PORT': '8080',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    result = runner(['apps', 'deploy', '--dry-run', '--template', _tmpl_path, 'RANDOM_LABEL=some_value'], env=env)

    assert result.exit_code == 0
    assert '--dry-run' in result.output


@responses.activate
def test_dry_run_fail(runner, json_fixture):
    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    result = runner(['apps', 'deploy', '--dry-run', '--template', _tmpl_path, 'RANDOM_LABEL=some_value'], env=env)

    assert not result.exit_code == 0


@responses.activate
def test_strategy_bluegreen_fails_without_marathon_lb(runner):
    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_DOCKER_EXPOSED_PORT': '8080',
        'SHPKPR_DEPLOYMENT_GROUP': 'test',
        'SHPKPR_DEPLOY_DOMAIN': 'test.com',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test-bluegreen.json.tmpl"
    result = runner(['apps', 'deploy', '--strategy', 'bluegreen', '--template', _tmpl_path], env=env)

    assert result.exit_code == 2
    assert 'Missing option "--marathon_lb_url".' in result.output


@responses.activate
def test_strategy_bluegreen(runner):
    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_MARATHON_LB_URL': "http://marathon-lb.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_DOCKER_EXPOSED_PORT': '8080',
        'SHPKPR_DEPLOYMENT_GROUP': 'test',
        'SHPKPR_DEPLOY_DOMAIN': 'test.com',
    }
    _tmpl_path = "tests/fixtures/templates/marathon/test-bluegreen.json.tmpl"
    result = runner(['apps', 'deploy', '--dry-run', '--strategy', 'bluegreen', '--template', _tmpl_path], env=env)

    assert result.exit_code == 0


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.wait')
def test_default_template(mock_deployment_wait, runner, json_fixture):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=201,
                  json=json_fixture("deployment"),
                  match_querystring=True)
    mock_deployment_wait.return_value = True

    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_MARATHON_APP_ID': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
    }
    result = runner(['apps', 'deploy'], env=env)

    assert result.exit_code == 0


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.wait')
def test_default_template_labels(mock_deployment_wait, runner, json_fixture):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=201,
                  json=json_fixture("deployment"),
                  match_querystring=True)
    mock_deployment_wait.return_value = True

    env = {
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_MARATHON_APP_ID': 'test-app',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
        'SHPKPR_LABEL_SOME_LABEL': 'somevalue',
    }
    result = runner(['apps', 'deploy'], env=env)

    assert result.exit_code == 0



# third-party imports
from cached_property import cached_property


class ApplicationStats(object):

    def __init__(self, app_definition, stats):
        self._app_definition = app_definition
        self._stats = stats

    @cached_property
    def listeners(self):
        """Returns all HAProxy listeners for the current application.
        """
        all_listeners = [r for r in self._stats if r.svname not in ["BACKEND", "FRONTEND"]]
        app_listeners = [l for l in all_listeners if l.pxname == self._deployment_label]
        return app_listeners

    def listeners_with_status(self, status):
        """Returns listeners with the given status string.
        """
        return [r for r in self.listeners if r.status == status.upper()]

    @property
    def listener_count(self):
        return len(self.listeners)

    @property
    def _deployment_label(self):
        """Construct and return a label in the format used by HAProxy to
        identify this application.

        This label can be used when filtering the list of HAProxy listeners to
        only those for this application.
        """
        labels = self._app_definition['labels']
        group = labels['HAPROXY_DEPLOYMENT_GROUP']
        port = labels['HAPROXY_0_PORT']
        return "{0}_{1}".format(group, port)



# third-party imports
import pytest

# local imports
from shpkpr.deployment.bluegreen.validate import AppDefinitionValidator
from shpkpr.deployment.bluegreen.validate import ValidationError


@pytest.fixture
def valid_app_definition(json_fixture):
    return json_fixture("marathon/bluegreen_app_new")


def test_valid_app(valid_app_definition):
    """Test that an app validates successfully when all the required labels are
    present.
    """
    validator = AppDefinitionValidator()
    try:
        validator.validate(valid_app_definition)
    except ValidationError as e:
        pytest.fail(e)


@pytest.fixture(params=[
    ["HAPROXY_DEPLOYMENT_GROUP"]
])
def invalid_app_definition(json_fixture, request):
    app_definition = json_fixture("marathon/bluegreen_app_new")
    for label in request.param:
        del(app_definition['labels'][label])
    return app_definition


def test_invalid_app(invalid_app_definition):
    """Test that an app fails validation when one or more required labels are
    missing.
    """
    validator = AppDefinitionValidator()
    with pytest.raises(ValidationError):
        validator.validate(invalid_app_definition)



# stdlib imports
import uuid

# third-party imports
import click


class HiddenOption(click.Option):
    """Option type that suppresses any help output.
    """
    hidden = True

    def get_help_record(self, ctx):
        return


def multioption(options, name=None, callback=None):
    """Attaches multiple options to a command function and optionally attaches a
    callback to a hidden option that's fired after parsing of the embedded
    options.
    """
    def decorator(f):
        if callback is not None:
            # this option exists mainly so that it can fire the callback and
            # that means we probably don't want users to invoke it on the
            # command line, accidentally or otherwise. Set its CLI flag to a
            # random(ish) string and hide it from the application's help output.
            f = click.option(
                "--{0}".format(uuid.uuid4().hex), name,
                callback=callback,
                cls=HiddenOption,
            )(f)
        # wrap the decorated function in all embedded decorators (in reverse
        # order so that they execute in the correct order at runtime).
        for option in reversed(options):
            f = option(f)
        return f
    return decorator



"""A collection of marathon-related utils
"""
# future imports
from __future__ import absolute_import

# third-party imports
import requests

# local imports
from .deployment import DeploymentNotFound
from .deployment import MarathonDeployment
from shpkpr import exceptions


class ClientError(exceptions.ShpkprException):
    pass


class DryRun(exceptions.ShpkprException):
    exit_code = 0


class MarathonClient(object):
    """A thin wrapper around marathon.MarathonClient for internal use
    """

    def __init__(self, marathon_url, username=None, password=None, dry_run=False):
        self._marathon_url = marathon_url
        self._dry_run = dry_run

        self._basic_auth = None
        if None not in [username, password]:
            self._basic_auth = requests.auth.HTTPBasicAuth(username, password)

    def _build_url(self, path):
        return self._marathon_url.rstrip("/") + path

    def _make_request(self, method, path, **kwargs):
        if self._dry_run:
            raise DryRun("Exiting as --dry-run requested")
        request = getattr(requests, method.lower())
        return request(self._build_url(path), auth=self._basic_auth, **kwargs)

    def embed_params(self, entity_type):
        return [
            "{0}.tasks".format(entity_type),
            "{0}.counts".format(entity_type),
            "{0}.deployments".format(entity_type),
            "{0}.lastTaskFailure".format(entity_type),
            "{0}.taskStats".format(entity_type),
        ]

    def get_info(self):
        """ Returns the marathon info from the /info endpoint
        """
        path = "/v2/info"
        response = self._make_request('GET', path)
        if response.status_code == 200:
            return response.json()
        raise ClientError("Unable to retrieve info from marathon")

    def delete_application(self, application_id, force=False):
        """Deletes the Application corresponding with application_id
        """
        path = "/v2/apps/" + application_id
        params = {"force": "true"} if force else {}
        response = self._make_request('DELETE', path, params=params)

        if response.status_code == 200:
            return True
        else:
            return False

    def get_application(self, application_id):
        """Returns detailed information for a single application.
        """
        path = "/v2/apps/" + application_id
        params = {"embed": self.embed_params("app")}
        response = self._make_request('GET', path, params=params)

        if response.status_code == 200:
            application = response.json()['app']
            return application

        # raise an appropriate error if something went wrong
        if response.status_code == 404:
            raise ClientError("Unable to retrieve application details from marathon: does not exist.")

        raise ClientError("Unknown Marathon error: %s\n\n%s" % (response.status_code, response.text))

    def list_applications(self):
        """Return a list of all applications currently deployed to marathon.
        """
        path = "/v2/apps"
        params = {"embed": self.embed_params("apps")}
        response = self._make_request('GET', path, params=params)

        if response.status_code == 200:
            applications = response.json()['apps']
            application_list = []
            for app in applications:
                application_list.append(app)
            return application_list

        raise ClientError("Unknown Marathon error: %s\n\n%s" % (response.status_code, response.text))

    def deploy(self, application_payload, force=False):
        """Deploys the given application(s) to Marathon.
        """
        # if the payload is a list and is one element long then we extract it
        # as we want to treat single app deploys differently. Doing this here
        # helps keep the cmd implementation clean.
        if isinstance(application_payload, (list, tuple)) and len(application_payload) == 1:
            application_payload = application_payload[0]

        # if at this point our payload is a dict then we treat it as a single
        # app, otherwise we treat it as a list of multiple applications to be
        # deployed together.
        if isinstance(application_payload, (list, tuple)):
            path = "/v2/apps/"
        else:
            path = "/v2/apps/" + application_payload['id']

        params = {"force": "true"} if force else {}
        response = self._make_request('PUT', path, params=params, json=application_payload)

        if response.status_code in [200, 201]:
            deployment = response.json()
            return MarathonDeployment(self, deployment['deploymentId'])

        # raise an appropriate error if something went wrong
        if response.status_code == 409:
            deployment_ids = ', '.join([x['id'] for x in response.json()['deployments']])
            raise ClientError("App(s) locked by one or more deployments: %s" % deployment_ids)

        raise ClientError("Unknown Marathon error: %s\n\n%s" % (response.status_code, response.text))

    def get_deployment(self, deployment_id):
        """Returns detailed information for a single deploy
        """
        response = self._make_request('GET', "/v2/deployments")

        if response.status_code == 200:
            for deployment in response.json():
                if deployment['id'] == deployment_id:
                    return deployment
            raise DeploymentNotFound(deployment_id)

        raise ClientError("Unknown Marathon error: %s\n\n%s" % (response.status_code, response.text))



# stdlib imports
import functools

# third-party imports
import pytest
from click.testing import CliRunner

# local imports
from shpkpr.cli.entrypoint import cli


@pytest.fixture
def runner():
    runner = CliRunner()
    return functools.partial(runner.invoke, cli)



# stdlib imports
import socket

# third-party imports
import six.moves.urllib as urllib


def resolve(marathon_lb_url):
    """Return the individual URLs for all available Marathon-LB instances given
    a single URL to a DNS-balanced Marathon-LB cluster.

    Marathon-LB typically uses DNS for load balancing between instances and so
    the address provided by the user may actually be multiple load-balanced
    instances. This function uses DNS to lookup the hostnames (IPv4 A-records)
    of each instance, returning them all to the caller for use as required.
    """
    url = urllib.parse.urlparse(marathon_lb_url)
    all_hosts = _get_alias_records(url.hostname)
    resolved_urls = _reassemble_urls(url, all_hosts)
    return resolved_urls


def _get_alias_records(hostname):
    """Return all IPv4 A records for a given hostname
    """
    return socket.gethostbyname_ex(hostname)[2]


def _reassemble_urls(url, hosts):
    return [_reassemble_url(url, host) for host in hosts]


def _reassemble_url(url, host):
    """Reassemble a stringified URL with the host replaced
    """
    netloc = "{0}:{1}".format(host, url.port)
    parts = (url[0], netloc, url[2], url[3], url[4], url[5])
    return urllib.parse.urlunparse(parts)



# stdlib imports
import logging
import sys

# third-party imports
import click
import docker

# local imports
from shpkpr.cli import arguments
from shpkpr.cli import options
from shpkpr.cli.entrypoint import CONTEXT_SETTINGS
from shpkpr.deployment import BlueGreenDeployment
from shpkpr.deployment import StandardDeployment
from shpkpr.template import load_values_from_environment
from shpkpr.template import render_json_template
from shpkpr.vault import resolve_secrets


logger = logging.getLogger(__name__)


def _validate_strategy_bluegreen(marathon_lb_client, **kw):
    if marathon_lb_client is None:
        msg = 'Missing option "--marathon_lb_url".'
        raise click.UsageError(msg)


STRATEGIES = {
    "standard": {
        "executor": StandardDeployment,
        "validator": lambda **kw: None,
        "default_template": "marathon/default/standard.json.tmpl",
    },
    "bluegreen": {
        "executor": BlueGreenDeployment,
        "validator": _validate_strategy_bluegreen,
        "default_template": "marathon/default/bluegreen.json.tmpl",
    },
}


@click.group('apps', context_settings=CONTEXT_SETTINGS)
def cli():
    """Manage Marathon applications.
    """


@cli.command(short_help='Deploy an application to Marathon.',
             context_settings=CONTEXT_SETTINGS)
@arguments.env_pairs
@options.force
@options.template_names
@options.template_path
@options.env_prefix
@options.timeout
@options.deployment_strategy
@options.marathon_lb_client
@options.marathon_client
def deploy(marathon_client, marathon_lb_client, deployment_strategy, timeout,
           env_prefix, template_path, template_names, force, env_pairs, **kw):
    """Deploy one or more applications to Marathon.
    """
    # select the appropriate deployment strategy
    strategy = STRATEGIES[deployment_strategy]

    # use the default template if none was specified
    if not template_names:
        template_names = [strategy["default_template"]]

    # read and render deploy template using values from the environment
    values = load_values_from_environment(prefix=env_prefix, overrides=env_pairs)
    rendered_templates = []
    for template_name in template_names:
        rendered_templates.append(render_json_template(template_path, template_name, **values))

    # perform some extra pre-flight validation if required, and execute the
    # deployment, blocking until complete.
    deployment_params = {"marathon_client": marathon_client,
                         "marathon_lb_client": marathon_lb_client,
                         "timeout": timeout,
                         "app_definitions": rendered_templates}
    strategy["validator"](**deployment_params)
    strategy["executor"](**deployment_params).execute(force)


@cli.command(short_help='Show application details.',
             context_settings=CONTEXT_SETTINGS)
@options.output_formatter
@options.application_id
@options.marathon_client
def show(marathon_client, application_id, output_formatter, **kw):
    """Show detailed information for one or more applications.
    """
    if application_id is None:
        payload = marathon_client.list_applications()
    else:
        payload = marathon_client.get_application(application_id)
    logger.info(output_formatter.format(payload))


@cli.command(short_help='Run a one-off task in a production-like environment.',
             context_settings=CONTEXT_SETTINGS)
@arguments.command
@arguments.env_pairs
@options.vault_client
@options.template_names
@options.template_path
@options.env_prefix
def run(env_prefix, template_path, template_names, vault_client, command, env_pairs, **kw):
    """Run a one-off task in a "production-like" environment.

    Uses the current deployment configuration to start up a container locally
    and run a single command. Environment variables and secrets are extracted
    from the current deployment configuration and injected into the container
    when starting up.
    """
    # use the default template if none was specified. We can just use the
    # standard template here since all we care about are environment variables
    # and secrets.
    if not template_names:
        template_name = STRATEGIES["standard"]["default_template"]
    else:
        template_name = template_names[0]

    # read and render deploy template using values from the environment
    values = load_values_from_environment(prefix=env_prefix, overrides=env_pairs)
    rendered_template = render_json_template(template_path, template_name, **values)

    # extract required configuration from the rendered template
    app_id = rendered_template["id"]
    docker_image = rendered_template["container"]["docker"]["image"]
    environment_variables = rendered_template["env"]
    environment_variables.update(resolve_secrets(vault_client, rendered_template))

    logger.info("Running command for app ({0}): {1}\n".format(app_id, command))

    docker_client = docker.from_env()
    container = docker_client.containers.run(docker_image,
                                             command,
                                             detach=True,
                                             auto_remove=True,
                                             environment=environment_variables)

    # stream logs from the running container to stdout
    for line in container.logs(stream=True):
        logger.info(line.decode('utf-8').rstrip())

    # exit with the container's exit code once it finishes
    sys.exit(container.wait()['StatusCode'])



# stdlib imports
import functools
import sys

# third-party imports
import six
from click.exceptions import ClickException


class ShpkprException(ClickException):
    """Common base class for all exceptions raised explicitly by shpkpr.

    Exceptions which are subclasses of this type will be handled nicely by
    shpkpr and a human-readable error message will be displayed to the user
    before exiting with an appropriate non-zero exit code. Any exceptions
    raised which are not a subclass of this type will exit(1) and print a
    traceback to the user's console.
    """
    pass


def _args_from_exception(exception):
    """In Python 3 we can't guarantee that an exception will have a `message`
    property. This function ensures that we can extract *args and have it
    include at least one argument, regardless of the exception type passed-in.
    """
    if exception.args:
        return exception.args
    if hasattr(exception, "message"):
        return (exception.message,)
    return (str(exception),)


def rewrap(exceptions_to_catch, exception_to_rewrap_with=ShpkprException):
    """Decorator that catches exceptions of type `exceptions_to_catch` (can be
    a list) and rewraps them with the given `exception_to_rewrap_with` type
    before re-raising with the original stack trace.
    """
    def real_decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except exceptions_to_catch:
                # the syntax for raising an exception with a traceback changed
                # in Python 3 and the python 2 version will cause the code to
                # throw a syntax error when parsed, so we use six to provide a
                # version agnostic `reraise` function we can use in all
                # environments.
                ei, tb = sys.exc_info()[1:]
                if six.PY2:
                    six.reraise(exception_to_rewrap_with, ei.message, tb)
                else:
                    ei_args = _args_from_exception(ei)
                    six.reraise(exception_to_rewrap_with, exception_to_rewrap_with(*ei_args), tb)
        return wrapper
    return real_decorator



# stdlib imports
import copy
import random
from datetime import datetime


def prepare_app_definition(app_definition, old_app_definition=None, apps_state=None, marathon_info=None):
    """Prepares ``app_definition`` for blue/green deployment by adding the
    necessary Marathon labels and setting the ID and service port.

    If an old app definition is passed in, it will be used to correctly pick the
    color and port for the new deployment, otherwise we default to "blue" and
    use the ports as defined in the new app definition.
    """
    # make a deep copy of the new app definition before we start mutating it so
    # we can refer back to properties of the original as needed.
    new_app_definition = copy.deepcopy(app_definition)

    if old_app_definition is not None:
        _rotate_colour(new_app_definition, old_app_definition)
        _rotate_port(new_app_definition, old_app_definition)
    else:
        _set_app_colour(new_app_definition, 'blue')
        _init_ports(new_app_definition, apps_state, marathon_info)

    new_app_definition['id'] = new_app_definition['id'] + '-' + _get_app_colour(new_app_definition)
    new_app_definition['labels']['HAPROXY_APP_ID'] = app_definition['id']
    new_app_definition['labels']['HAPROXY_DEPLOYMENT_STARTED_AT'] = datetime.now().isoformat()
    new_app_definition['labels']['HAPROXY_DEPLOYMENT_TARGET_INSTANCES'] = str(new_app_definition['instances'])
    return new_app_definition


def _init_ports(app_definition, apps_state, marathon_info):
    """Determines that start port for the application and updates the application state.
    This works for both Dockerised and non-Dockerised applications.
    """
    if _is_service_port_defined(app_definition):
        service_port = _get_service_port(app_definition)
        alt_port = _get_alt_port(app_definition)
    else:
        service_port = _select_next_port(apps_state, marathon_info, [])
        alt_port = _select_next_port(apps_state, marathon_info, [service_port])

    app_definition = _set_current_port(app_definition, service_port)
    app_definition = _set_service_port(app_definition, service_port)
    app_definition = _set_alt_port(app_definition, alt_port)
    return app_definition


def _select_next_port(apps_state, marathon_info, used_ports):
    """Given a currently deployed app definition, return the service port to use
    for the next blue/green deployment.
    """
    min_port = marathon_info['marathon_config']['local_port_min']
    max_port = marathon_info['marathon_config']['local_port_max']

    for app_definition in apps_state:
        used_ports.append(_get_service_port(app_definition))
        used_ports.append(_get_alt_port(app_definition))

    while True:
        port = random.randint(min_port, max_port)
        if port not in used_ports:
            return port


def _rotate_port(new_app_definition, old_app_definition):
    """Given a currently deployed app definition, return the service port to use
    for the next blue/green deployment.
    """
    primary_port = _get_service_port(old_app_definition)
    secondary_port = _get_alt_port(old_app_definition)
    current_port = _get_current_port(old_app_definition)
    if current_port == primary_port:
        new_app_definition = _set_current_port(new_app_definition, secondary_port)
    else:
        new_app_definition = _set_current_port(new_app_definition, primary_port)

    _set_service_port(new_app_definition, _get_service_port(old_app_definition))
    _set_alt_port(new_app_definition, _get_alt_port(old_app_definition))
    return new_app_definition


def _rotate_colour(new_app_definition, old_app_definition):
    """Given a currently deployed app definition, return the "color" to use for
    the next blue/green deployment.

    Right now this simply swaps the colors used, but could be extended in future
    to allow for more complex behaviour.
    """
    colours = {"blue": "green", "green": "blue"}
    old_colour = _get_app_colour(old_app_definition)
    return _set_app_colour(new_app_definition, colours[old_colour])


def _is_service_port_defined(app_definition):
    return _get_service_port(app_definition) is not None and 'HAPROXY_DEPLOYMENT_ALT_PORT' in app_definition['labels']


def _set_current_port(app_definition, service_port):
    """Set the service port on the provided app definition.

    This works for both Dockerised and non-Dockerised applications.
    """
    try:
        port_mappings = app_definition['container']['docker']['portMappings']
        port_mappings[0]['servicePort'] = service_port
    except (KeyError, IndexError):
        app_definition['ports'][0] = service_port

    return app_definition


def _set_service_port(app_definition, service_port):
    """Set the service port on the provided app definition.

    This works for both Dockerised and non-Dockerised applications.
    """
    app_definition['labels']['HAPROXY_0_PORT'] = str(service_port)
    return app_definition


def _set_alt_port(app_definition, alt_port):
    """Set the alt port on the provided app definition.

    This works for both Dockerised and non-Dockerised applications.
    """
    app_definition['labels']['HAPROXY_DEPLOYMENT_ALT_PORT'] = str(alt_port)
    return app_definition


def _get_current_port(app_definition):
    try:
        port_mappings = app_definition['container']['docker']['portMappings']
        service_port = port_mappings[0]['servicePort']
    except (KeyError, IndexError):
        if 'ports' in app_definition:
            service_port = app_definition['ports'][0]
        else:
            return None
    return int(service_port)


def _get_alt_port(app_definition):
    if 'HAPROXY_DEPLOYMENT_ALT_PORT' in app_definition['labels']:
        return int(app_definition['labels']['HAPROXY_DEPLOYMENT_ALT_PORT'])
    return None


def _get_service_port(app_definition):
    if 'HAPROXY_0_PORT' in app_definition['labels']:
        return int(app_definition['labels']['HAPROXY_0_PORT'])
    return _get_current_port(app_definition)


def _get_app_colour(app_definition):
    return app_definition['labels']['HAPROXY_DEPLOYMENT_COLOUR']


def _set_app_colour(app_definition, colour):
    app_definition['labels']['HAPROXY_DEPLOYMENT_COLOUR'] = colour
    return app_definition



# third-party imports
import pytest
from freezegun import freeze_time

# local imports
from shpkpr.deployment.bluegreen.prepare import prepare_app_definition


@pytest.fixture
@freeze_time("2011-11-11T11:11:11.111111Z")
def app_definition(json_fixture):
    """Prepare an app definition for deployment.

    This fixture prepares an app and assumes that no existing stack is present
    on Marathon.
    """
    app_definition_fixture = json_fixture("marathon/bluegreen_app_new")
    app_states_fixture = json_fixture("valid_apps")
    return prepare_app_definition(app_definition_fixture, None, app_states_fixture['apps'])


def test_color(app_definition):
    """Test that the app's color is set as a label
    """
    actual = app_definition['labels']['HAPROXY_DEPLOYMENT_COLOUR']
    expected = "blue"

    assert actual == expected


def test_id(app_definition):
    """Test that the app's ID is correctly transformed
    """
    actual = app_definition['id']
    expected = "my-app-blue"

    assert actual == expected


def test_docker_service_port(app_definition):
    """Test that the service port is set correctly when using docker
    """
    actual = app_definition['container']['docker']['portMappings'][0]['servicePort']
    expected = 11090

    assert actual == expected


def test_labels_id(app_definition):
    """Test that the app's original ID is preserved as a label
    """
    actual = app_definition['labels']['HAPROXY_APP_ID']
    expected = "my-app"

    assert actual == expected


def test_labels_port(app_definition):
    """Test that the app's original port is preserved as a label
    """
    actual = app_definition['labels']['HAPROXY_0_PORT']
    expected = "11090"

    assert actual == expected


def test_labels_alt_port(app_definition):
    """Test that the app's original port is preserved as a label
    """
    actual = app_definition['labels']['HAPROXY_DEPLOYMENT_ALT_PORT']
    expected = "11091"

    assert actual == expected


def test_labels_deployment_started_at(app_definition):
    """Test that a timestamp is correctly added to the app's labels
    """
    actual = app_definition['labels']['HAPROXY_DEPLOYMENT_STARTED_AT']
    expected = "2011-11-11T11:11:11.111111"

    assert actual == expected



# third-party imports
import pytest

# local imports
from shpkpr import template_filters


def test_filter_items():
    values = template_filters.filter_items({"X": "a", "Y": "b", "Z": "c"})
    assert ("X", "a") in values
    assert ("Y", "b") in values
    assert ("Z", "c") in values


def test_filter_items_with_startswith():
    values = template_filters.filter_items({"_X": "a", "_Y": "b", "Z": "c"}, startswith="_")
    assert ("_X", "a") in values
    assert ("_Y", "b") in values
    assert ("Z", "c") not in values


def test_filter_items_with_startswith_strip_prefix():
    values = template_filters.filter_items({"_X": "a", "_Y": "b", "Z": "c"}, startswith="_", strip_prefix=True)
    assert ("X", "a") in values
    assert ("Y", "b") in values
    assert ("Z", "c") not in values


def test_filter_items_with_strip_prefix():
    values = template_filters.filter_items({"_X": "a", "_Y": "b", "Z": "c"}, strip_prefix=True)
    assert ("_X", "a") in values
    assert ("_Y", "b") in values
    assert ("Z", "c") in values


def test_require_int():
    i = template_filters.require_int("1")
    assert i == 1


def test_require_int_in_range():
    i = template_filters.require_int("1", min=0, max=5)
    assert i == 1


def test_require_int_not_an_int():
    with pytest.raises(template_filters.IntegerRequired):
        template_filters.require_int("abc")


def test_require_int_too_low():
    with pytest.raises(template_filters.IntegerTooSmall):
        template_filters.require_int("-1", min=0)


def test_require_int_too_high():
    with pytest.raises(template_filters.IntegerTooLarge):
        template_filters.require_int("1", max=0)


def test_require_float():
    i = template_filters.require_float("1")
    assert i == 1


def test_require_float_in_range():
    i = template_filters.require_float("1", min=0, max=5)
    assert i == 1


def test_require_float_not_an_float():
    with pytest.raises(template_filters.FloatRequired):
        template_filters.require_float("abc")


def test_require_float_too_low():
    with pytest.raises(template_filters.FloatTooSmall):
        template_filters.require_float("-1", min=0)


def test_require_float_too_high():
    with pytest.raises(template_filters.FloatTooLarge):
        template_filters.require_float("1", max=0)



# stdlib imports
import json

# third-party imports
import yaml


class OutputFormatter(object):
    """OutputFormatter is used to serialise complex Python objects for printing
    to the log.
    """

    def __init__(self, fmt):
        self.fmt = fmt

    def format(self, obj):
        formatters = {
            "json": self.to_json,
            "yaml": self.to_yaml,
        }
        return formatters[self.fmt](obj)

    def to_json(self, obj):
        return json.dumps(obj, indent=4, sort_keys=True)

    def to_yaml(self, obj):
        return yaml.dump(obj, default_flow_style=False)



# stdlib imports
import os
import sys

# third-party imports
import click

# local imports
from . import logger


CONTEXT_SETTINGS = dict(auto_envvar_prefix='SHPKPR')


cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, 'commands'))


class ShpkprCLI(click.MultiCommand):

    def list_commands(self, ctx):
        files = [f for f in os.listdir(cmd_folder) if f.endswith('.py')]
        return sorted([f[4:-3] for f in files if f.startswith('cmd_')])

    def get_command(self, ctx, name):
        try:
            if sys.version_info[0] == 2:
                name = name.encode('ascii', 'replace')
            mod = __import__('shpkpr.commands.cmd_' + name, None, None, ['cli'])
        except ImportError:
            return
        return mod.cli


@click.command(cls=ShpkprCLI, context_settings=CONTEXT_SETTINGS)
def cli():
    """A tool to manage applications running on Marathon."""
    logger.configure()



# stdlib imports
import logging


logger = logging.getLogger(__name__)


class StandardDeployment(object):
    """StandardDeployment implements Marathon's basic deployment workflow and
    uses the primitives provided by the Marathon API to perform a standard
    rolling deploy according to application settings.

    This deployment strategy is best suited for non-web-facing applications or
    those that can tolerate minor downtime during deployment e.g. consumer or
    worker applications.
    """

    def __init__(self, marathon_client, timeout, app_definitions, **kw):
        self.marathon_client = marathon_client
        self.timeout = timeout
        self.app_definitions = app_definitions

    def execute(self, force=False):
        """Execute standard Marathon deployment.
        """
        app_ids = ", ".join([a["id"] for a in self.app_definitions])
        logger.info("Executing standard deployment: {0}".format(app_ids))

        deployment = self.marathon_client.deploy(
            self.app_definitions,
            force=force,
        )

        logger.info("Waiting for marathon deployment to complete: {0}".format(deployment.deployment_id))
        result = deployment.wait(timeout=self.timeout)
        logger.info("Marathon deployment complete: {0}".format(deployment.deployment_id))
        return result



# stdlib imports
import collections

# third-party imports
import pytest

# local imports
from shpkpr.marathon_lb.stats import Stats


@pytest.fixture
def haproxy_stats(file_fixture):
    csv_data = file_fixture("haproxy/stats.csv")
    return Stats(csv_data)


def test_stats_is_iterable(haproxy_stats):
    assert isinstance(haproxy_stats, collections.Iterable)


def test_stats_can_be_indexed(haproxy_stats):
    try:
        haproxy_stats[0]
        haproxy_stats[-1]
    except IndexError as e:
        pytest.fail("Unable to index `Stats` instance: {0}".format(e))


def test_stats_have_a_length(haproxy_stats):
    try:
        len(haproxy_stats)
    except TypeError as e:
        pytest.fail("`Stats` instance does not have a length: {0}".format(e))


def test_stats_does_not_contain_commented_rows(haproxy_stats):
    for stat in haproxy_stats:
        assert not stat[0].startswith("#")


def test_parse_haproxy_stats_count(haproxy_stats):
    assert len(haproxy_stats) == 31


def test_parse_haproxy_stats_first_row(haproxy_stats):
    assert haproxy_stats[0].pxname == 'http-in'
    assert haproxy_stats[0].svname == 'FRONTEND'


def test_parse_haproxy_stats_last_row(haproxy_stats):
    assert haproxy_stats[-1].pxname == 'http-in'
    assert haproxy_stats[-1].svname == '10_0_6_25_23336'



# stdlib imports
import logging


logger = logging.getLogger(__name__)


def resolve_secrets(vault_client, rendered_template):
    """Parse a rendered template, extract any secret definitions, retrieve them
    from vault and return them to the caller.

    This is used in situations where direct Vault support is not available e.g.
    Chronos.
    """
    resolved_secrets = {}

    secrets = rendered_template.get("secrets", {})
    for name, definition in secrets.items():
        # parse the secret source and retrieve from vault
        path, key = definition["source"].split(":")
        secret = vault_client.read(path)
        if secret:
            resolved_secrets[name] = secret["data"][key]
        else:
            logger.info("Couldn't locate secret in Vault: {0}".format(path))

    return resolved_secrets



# third-party imports
import mock
import pytest

# local imports
from shpkpr.deployment.bluegreen.validate import MarathonStateValidator
from shpkpr.deployment.bluegreen.validate import ValidationError


@pytest.fixture
def valid_app_definition(json_fixture):
    return json_fixture("marathon/bluegreen_app_new")


def test_valid_state(valid_app_definition):
    """Test that an app validates successfully when no existing stacks are
    present on Marathon.
    """
    marathon_client = mock.Mock()
    # we use a lambda here because mock interprets the empty list as list of
    # return values or exceptions to be raised and a single empty list causes an
    # error. the two options are wrapping the empty list in an outer list, or a
    # function that returns the empty list.
    marathon_client.list_applications.side_effect = lambda: []

    validator = MarathonStateValidator(marathon_client)
    try:
        validator.validate(valid_app_definition)
    except ValidationError as e:
        pytest.fail(e)


def test_valid_state_one_existing_stack(valid_app_definition, json_fixture):
    """Test that an app validates successfully when one existing stack is
    present on Marathon.
    """
    marathon_client = mock.Mock()
    marathon_client.list_applications.side_effect = lambda: [
        json_fixture("marathon/bluegreen_app_existing"),
    ]

    validator = MarathonStateValidator(marathon_client)
    try:
        validator.validate(valid_app_definition)
    except ValidationError as e:
        pytest.fail(e)


def test_invalid_state_two_existing_stacks(valid_app_definition, json_fixture):
    """Test that an app fails validation when more than one existing stack is
    present on Marathon.
    """
    marathon_client = mock.Mock()
    marathon_client.list_applications.side_effect = lambda: [
        json_fixture("marathon/bluegreen_app_existing"),
        json_fixture("marathon/bluegreen_app_existing"),
    ]

    validator = MarathonStateValidator(marathon_client)
    with pytest.raises(ValidationError):
        validator.validate(valid_app_definition)



"""Argument parameter definitions for shpkpr.

All argument parameters should be defined in this module, commands should not
define their arguments directly, instead all arguments should be defined here
and imported as required. This helps maintain consistency where a single
argument is used for multiple commands.
"""
# third-party imports
import click


def _env_pairs_to_dict(ctx, param, value):
    """Converts a space-seperated list of key=value pairs into a Python
    dictionary.
    """
    d = {}
    for pair in [x.split('=', 1) for x in value]:
        if len(pair) == 1:
            d[pair[0]] = ''
        else:
            d[pair[0]] = pair[1]
    return d


command = click.argument(
    'command',
    nargs=1,
)

env_pairs = click.argument(
    'env_pairs',
    nargs=-1,
    callback=_env_pairs_to_dict,
)

job_name = click.argument(
    'job_name',
    nargs=1,
)



# third-party imports
import mock


def test_no_args(runner):
    result = runner(['cron'])

    assert result.exit_code == 0
    assert 'Usage:' in result.output


def test_help(runner):
    result = runner(['cron', '--help'])

    assert result.exit_code == 0
    assert 'Usage:' in result.output
    assert 'Manage Chronos jobs' in result.output


@mock.patch("shpkpr.cli.options.ChronosClient.list")
def test_show(mock_chronos_client, runner, json_fixture):
    mock_chronos_client.return_value = json_fixture("chronos_jobs")

    result = runner(['cron', 'show'], env={
        'SHPKPR_CHRONOS_URL': "chronos.somedomain.com:4400",
    })

    assert '"name": "foo-job"' in result.output
    assert '"name": "bar-job"' in result.output
    assert result.exit_code == 0


@mock.patch("shpkpr.cli.options.ChronosClient.list")
def test_show_job_name(mock_chronos_client, runner, json_fixture):
    mock_chronos_client.return_value = json_fixture("chronos_jobs")

    result = runner(['cron', 'show', '--job-name', 'foo-job'], env={
        'SHPKPR_CHRONOS_URL': "chronos.somedomain.com:4400",
    })

    assert '"name": "foo-job"' in result.output
    assert '"name": "bar-job"' not in result.output
    assert result.exit_code == 0


@mock.patch("shpkpr.cli.options.ChronosClient.list")
@mock.patch("shpkpr.cli.options.ChronosClient.add")
def test_set(mock_chronos_add, mock_chronos_list, runner, json_fixture):
    mock_chronos_list.return_value = []
    mock_chronos_add.return_value = True

    _tmpl_path = 'tests/fixtures/templates/chronos/test-chronos.json.tmpl'
    result = runner(['cron', 'set', '--template', _tmpl_path], env={
        'SHPKPR_CHRONOS_URL': "chronos.somedomain.com:4400",
        'SHPKPR_CHRONOS_JOB_NAME': 'shpkpr-test-job',
    })

    assert mock_chronos_add.called
    assert result.exit_code == 0


@mock.patch("shpkpr.cli.options.ChronosClient.list")
@mock.patch("shpkpr.cli.options.ChronosClient.add")
def test_set_default_template(mock_chronos_add, mock_chronos_list, runner, json_fixture):
    mock_chronos_list.return_value = []
    mock_chronos_add.return_value = True

    result = runner(['cron', 'set'], env={
        'SHPKPR_CHRONOS_URL': "chronos.somedomain.com:4400",
        'SHPKPR_CHRONOS_JOB_NAME': 'shpkpr-test-job',
        'SHPKPR_CHRONOS_OWNER': 'shpkpr-test@example.com',
        'SHPKPR_CHRONOS_CMD': 'someprogram --run',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
    })

    assert mock_chronos_add.called
    assert result.exit_code == 0


@mock.patch("shpkpr.cli.options.ChronosClient.list")
@mock.patch("shpkpr.cli.options.ChronosClient.add")
def test_set_default_template_labels(mock_chronos_add, mock_chronos_list, runner, json_fixture):
    mock_chronos_list.return_value = []
    mock_chronos_add.return_value = True

    result = runner(['cron', 'set'], env={
        'SHPKPR_CHRONOS_URL': "chronos.somedomain.com:4400",
        'SHPKPR_CHRONOS_JOB_NAME': 'shpkpr-test-job',
        'SHPKPR_CHRONOS_OWNER': 'shpkpr-test@example.com',
        'SHPKPR_CHRONOS_CMD': 'someprogram --run',
        'SHPKPR_LABEL_SOME_LABEL': 'some-value',
        'SHPKPR_DOCKER_REPOTAG': 'goexample/outyet:latest',
    })

    assert mock_chronos_add.called
    assert result.exit_code == 0


@mock.patch("shpkpr.cli.options.ChronosClient.list")
@mock.patch("shpkpr.cli.options.ChronosClient.add")
def test_set_multiple(mock_chronos_add, mock_chronos_list, runner):
    mock_chronos_list.return_value = []
    mock_chronos_add.return_value = True

    _tmpl_path = 'tests/fixtures/templates/chronos/test-chronos.json.tmpl'
    _tmpl_path_2 = 'tests/fixtures/templates/chronos/test-chronos-2.json.tmpl'
    result = runner(
        [
            'cron', 'set',
            '--template', _tmpl_path,
            '--template', _tmpl_path_2,
        ],
        env={
            'SHPKPR_CHRONOS_URL': "chronos.somedomain.com:4400",
            'SHPKPR_CHRONOS_JOB_NAME': 'shpkpr-test-job',
            'SHPKPR_CHRONOS_JOB_2_NAME': 'shpkpr-test-job-2',
        },
    )

    assert mock_chronos_add.called_twice
    assert result.exit_code == 0


@mock.patch("shpkpr.cli.options.ChronosClient.list")
@mock.patch("shpkpr.cli.options.ChronosClient.update")
@mock.patch("shpkpr.cli.options.ChronosClient.add")
def test_set_update(mock_chronos_add, mock_chronos_update, mock_chronos_list, runner):
    mock_chronos_list.return_value = [{'name': 'shpkpr-test-job'}]
    mock_chronos_update.return_value = True

    _tmpl_path = 'tests/fixtures/templates/chronos/test-chronos.json.tmpl'
    result = runner(['cron', 'set', '--template', _tmpl_path], env={
        'SHPKPR_CHRONOS_URL': "chronos.somedomain.com:4400",
        'SHPKPR_CHRONOS_JOB_NAME': 'shpkpr-test-job',
    })

    mock_chronos_add.assert_not_called()

    assert mock_chronos_update.called
    assert result.exit_code == 0


@mock.patch("shpkpr.cli.options.ChronosClient.delete")
def test_delete(mock_chronos_client, runner):
    mock_chronos_client.return_value = True

    result = runner(["cron", "delete", "test-job"], env={
        'SHPKPR_CHRONOS_URL': "chronos.somedomain.com:4400",
    })

    mock_chronos_client.assert_called_with('test-job')
    assert result.exit_code == 0


@mock.patch("shpkpr.cli.options.ChronosClient.delete_tasks")
def test_delete_tasks(mock_chronos_client, runner):
    mock_chronos_client.return_value = True

    result = runner(["cron", "delete-tasks", "test-job"], env={
        'SHPKPR_CHRONOS_URL': "chronos.somedomain.com:4400",
    })

    mock_chronos_client.assert_called_with('test-job')
    assert result.exit_code == 0


@mock.patch("shpkpr.cli.options.ChronosClient.run")
def test_run(mock_chronos_client, runner):
    mock_chronos_client.return_value = True

    result = runner(["cron", "run", "test-job"], env={
        'SHPKPR_CHRONOS_URL': "chronos.somedomain.com:4400",
    })

    mock_chronos_client.assert_called_with('test-job')
    assert result.exit_code == 0



# local imports
from .client import ClientError
from .client import DryRun
from .client import MarathonClient
from .deployment import DeploymentFailed
from .deployment import DeploymentNotFound
from .deployment import MarathonDeployment


__all__ = [
    'ClientError',
    'DeploymentFailed',
    'DeploymentNotFound',
    'DryRun',
    'MarathonClient',
    'MarathonDeployment',
]



# stdlib imports
import csv
from collections import namedtuple

# third-party imports
from cached_property import cached_property


class Stats(object):

    def __init__(self, *instance_stats):
        self._raw_stats = instance_stats

    def __iter__(self):
        return (r for r in self._stats)

    def __getitem__(self, i):
        return self._stats[i]

    def __len__(self):
        return len(self._stats)

    @cached_property
    def _stats(self):
        """Parse and aggregate all stats files, caching the result
        """
        _stats = []
        for instance_stats in self._raw_stats:
            _stats.extend(self._parse_instance_stats(instance_stats))
        return _stats

    def _parse_instance_stats(self, instance_stats):
        """Parse a single HAProxy stats CSV file.
        """
        rows = instance_stats.splitlines()
        headers = rows[0].lstrip('# ').rstrip(',\n').split(',')
        csv_reader = csv.reader(rows[1:], quotechar="'")

        Row = namedtuple('Row', headers)
        return [Row(*row[0:-1]) for row in csv_reader if row[0][0] != '#']



# stdlib imports
import logging
import time

# local imports
from .prepare import prepare_app_definition
from .validate import Validator
from .wait import SwapApplicationTimeout
from .wait import Waiter
from shpkpr.marathon import DeploymentFailed
from cached_property import cached_property

logger = logging.getLogger(__name__)


class BlueGreenDeployment(object):
    """BluegreenDeployment implements a blue-green deployment strategy for
    web-facing applications on Marathon.

    It builds on the primitives provided by Marathon to provide zero-downtime
    deployments for HTTP services exposed to the web via Marathon-LB.
    """

    def __init__(self, marathon_client, marathon_lb_client, timeout, app_definitions, **kw):
        self.marathon_client = marathon_client
        self.marathon_lb_client = marathon_lb_client
        self.timeout = timeout
        self.app_definitions = app_definitions

    def execute(self, force=False):
        """Execute a bluegreen deployment.

        The mechanics of a bluegreen deploy mean that we have to deploy each
        application in turn (instead of the all-at-once approach used for
        standard deployments).

        NOTE: We could maybe be smarter here in future e.g. perhaps don't tear
        down old stacks until _all_ new stacks are running so we can roll back
        partially successful invocations.
        """
        app_ids = ", ".join([a["id"] for a in self.app_definitions])
        logger.info("Executing bluegreen deployment: {0}".format(app_ids))

        for app_definition in self.app_definitions:
            self.execute_for_app(app_definition, force)

    def execute_for_app(self, app_definition, force):
        """Execute a bluegreen deployment for a single application.
        """
        deployment = BlueGreenDeploymentSingleApp(self.marathon_client,
                                                  self.marathon_lb_client,
                                                  self.timeout,
                                                  app_definition)
        return deployment.execute(force=force)


class BlueGreenDeploymentSingleApp(object):

    def __init__(self, marathon_client, marathon_lb_client, timeout, app_definition):
        self.marathon_client = marathon_client
        self.marathon_lb_client = marathon_lb_client
        self.timeout = timeout
        self.app_definition = app_definition

    def execute(self, force=False):
        """Execute a bluegreen deployment for a single application.
        """
        Validator(self.marathon_client).validate(self.app_definition)

        # prepare the new application for deployment by setting the appropriate
        # labels and transforming the app's ID as appropriate. These changes
        # will allow Marathon-LB to cut the traffic over to the new app once
        # deployed.
        apps_state = self._fetch_apps_state
        marathon_info = self._fetch_marathon_info
        old_app_definition = self._fetch_old_app_definition()
        new_app_definition = prepare_app_definition(self.app_definition, old_app_definition, apps_state, marathon_info)

        # deploy the new application, wait until the traffic is cut over and
        # then tear down the old stack. If the process fails for any reason then
        # we roll back by tearing down the newly deployed stack at which point
        # Marathon-LB will begin serving traffic to the old application again
        # (if it had already stopped).
        old_app_id = old_app_definition["id"] if old_app_definition else None
        new_app_id = new_app_definition["id"]
        try:
            self._deploy_new_application(new_app_definition)
            self._wait_for_traffic_swap(old_app_id, new_app_id)
        except (DeploymentFailed, SwapApplicationTimeout):
            self._remove_new_application(new_app_id)
        else:
            self._remove_old_application(old_app_id)

    def _deploy_new_application(self, new_app_definition):
        """Deploy a new application definition to Marathon.
        """
        deployment = self.marathon_client.deploy([new_app_definition])

        logger.info("Waiting for marathon deployment to complete: {0}".format(deployment.deployment_id))
        result = deployment.wait(timeout=self.timeout)
        logger.info("Marathon deployment complete: {0}".format(deployment.deployment_id))
        return result

    def _wait_for_traffic_swap(self, old_app_id, new_app_id):
        """Wait for traffic to cut over from the old to the new application
        """
        if old_app_id is None:
            return

        waiter = Waiter(self.marathon_client,
                        self.marathon_lb_client,
                        new_app_id,
                        old_app_id)

        deadline = time.time() + self.timeout
        return waiter.wait(deadline)

    def _remove_application(self, app_id):
        """Remove a deployed application from Marathon
        """
        success = self.marathon_client.delete_application(app_id)
        if success:
            logger.info("Successfully removed application: %s", app_id)
        else:
            logger.info("Unable to remove application, manual intervention required: %s", app_id)
        return success

    def _remove_new_application(self, app_id):
        """Removes the newly started application after a deploy error
        """
        logger.info("Deployment failed: removing newly deployed stack")
        self._remove_application(app_id)
        raise DeploymentFailed("Deployment failed")

    def _remove_old_application(self, app_id):
        """Removes the old application after a successful deploy
        """
        if app_id is None:
            return

        logger.info("About to delete old application: %s", app_id)
        if not self._remove_application(app_id):
            raise DeploymentFailed("Deployment failed")

    def _fetch_old_app_definition(self):
        """Fetches the application definition for the currently deployed stack
        from Marathon if one exists. If no existing stacks are found then
        ``None`` is returned instead.
        """
        def get_deployment_group(app_definition):
            labels = app_definition.get('labels', {})
            return labels.get('HAPROXY_DEPLOYMENT_GROUP')

        target_deployment_group = get_deployment_group(self.app_definition)
        for app_definition in self._fetch_apps_state:
            if get_deployment_group(app_definition) == target_deployment_group:
                return app_definition
        return None

    @cached_property
    def _fetch_marathon_info(self):
        """Fetch current marathon configuration info
        """
        return self.marathon_client.get_info()

    @cached_property
    def _fetch_apps_state(self):
        """Fetch current marathon app state
        """
        return self.marathon_client.list_applications()



# third-party imports
import mock
import pytest

# local imports
from shpkpr.marathon import DeploymentFailed
from shpkpr.marathon import DeploymentNotFound
from shpkpr.marathon import MarathonClient
from shpkpr.marathon import MarathonDeployment


@mock.patch('shpkpr.marathon.MarathonDeployment.check')
def test_deployment_wait(mock_deployment_check):
    """ Test that deployment.wait() spins until a deployment succeeds.
    """
    mock_deployment_check.side_effect = [False, False, False, False, False,
                                         False, False, False, False, False,
                                         False, False, False, False, False,
                                         False, False, False, False, True]

    client = MarathonClient("http://marathon.somedomain.com:8080")
    deployment = MarathonDeployment(client, '1234')
    deployment.wait(check_interval_secs=0.01)
    assert mock_deployment_check.call_count == 20


@mock.patch('shpkpr.marathon.MarathonDeployment.check')
def test_deployment_wait_with_timeout(mock_deployment_check):
    """ Test that deployment.wait() spins until a deployment succeeds.
    """
    mock_deployment_check.side_effect = [False, False, False]

    client = MarathonClient("http://marathon.somedomain.com:8080")
    deployment = MarathonDeployment(client, '1234')
    with pytest.raises(DeploymentFailed):
        deployment.wait(timeout=1, check_interval_secs=0.5)
    assert mock_deployment_check.call_count == 2


@mock.patch('shpkpr.marathon.MarathonClient.get_deployment')
def test_deployment_check(mock_get_deployment):
    """ Test that deployment.check() returns True when the app has deployed successfully.
    """
    mock_get_deployment.side_effect = DeploymentNotFound("1234")

    client = MarathonClient("http://marathon.somedomian.com:8080")
    deployment = MarathonDeployment(client, '1234')
    assert deployment.check()


@mock.patch('shpkpr.marathon.MarathonClient.get_deployment')
def test_deployment_check_with_deployment_in_progress(mock_get_deployment):
    """ Test that deployment.check() returns False when the app has a deployment in progress.
    """
    deploy_in_progress = {
        "id": "97c136bf-5a28-4821-9d94-480d9fbb01c8",
        "version": "2015-09-30T09:09:17.614Z",
        "affectedApps": ["/foo"],
        "steps": [
            [
                {
                    "action": "StartApplication",
                    "app": "/foo"
                }
            ],
            [
                {
                    "action": "ScaleApplication",
                    "app": "/foo"
                }
            ]
        ],
        "currentActions": [
            {
                "action": "ScaleApplication",
                "app": "/foo"
            }
        ],
        "currentStep": 2,
        "totalSteps": 2
    }
    mock_get_deployment.return_value = deploy_in_progress

    client = MarathonClient("http://marathon.somedomian.com:8080")
    deployment = MarathonDeployment(client, "97c136bf-5a28-4821-9d94-480d9fbb01c8")
    assert not deployment.check()



# third-party imports
import pytest

# local imports
from shpkpr.template import InvalidJSONError
from shpkpr.template import MissingTemplateError
from shpkpr.template import UndefinedError
from shpkpr.template import load_values_from_environment
from shpkpr.template import render_json_template
from shpkpr.template_filters import IntegerRequired
from shpkpr.template_filters import IntegerTooLarge
from shpkpr.template_filters import IntegerTooSmall
from shpkpr.template_filters import FloatRequired
from shpkpr.template_filters import FloatTooLarge
from shpkpr.template_filters import FloatTooSmall


def _write_template_to_disk(tmpdir, template_name, template_data):
    """shpkpr loads template files from disk normally. This convenience
    function writes a template file to disk and returns a (directory, name)
    tuple.
    """
    with tmpdir.join(template_name).open("w") as f:
        f.write(template_data)
    return (tmpdir.strpath, template_name)


def test_load_environment_vars_without_prefix(monkeypatch):
    monkeypatch.setenv('BANANA', 'bread')
    monkeypatch.setenv('STRAWBERRY', 'cheesecake')
    monkeypatch.setenv('APPLE_AND_BLACKCURRANT', 'crumble')

    values = load_values_from_environment()

    assert 'BANANA' in values
    assert values['BANANA'] == 'bread'
    assert 'STRAWBERRY' in values
    assert values['STRAWBERRY'] == 'cheesecake'
    assert 'APPLE_AND_BLACKCURRANT' in values
    assert values['APPLE_AND_BLACKCURRANT'] == 'crumble'


def test_load_environment_vars_with_prefix(monkeypatch):
    monkeypatch.setenv('BANANA', 'bread')
    monkeypatch.setenv('SHPKPR_STRAWBERRY', 'cheesecake')
    monkeypatch.setenv('SHPKPR_APPLE_AND_BLACKCURRANT', 'crumble')
    monkeypatch.setenv('SHPKPR_SHPKPR_APPLE_AND_BLACKCURRANT', 'crumble')

    values = load_values_from_environment("SHPKPR")

    assert 'BANANA' not in values
    assert 'STRAWBERRY' in values
    assert values['STRAWBERRY'] == 'cheesecake'
    assert 'APPLE_AND_BLACKCURRANT' in values
    assert values['APPLE_AND_BLACKCURRANT'] == 'crumble'
    assert 'SHPKPR_APPLE_AND_BLACKCURRANT' in values
    assert values['SHPKPR_APPLE_AND_BLACKCURRANT'] == 'crumble'


def test_load_environment_vars_with_prefix_with_trailing_underscore(monkeypatch):
    monkeypatch.setenv('BANANA', 'bread')
    monkeypatch.setenv('SHPKPR_STRAWBERRY', 'cheesecake')
    monkeypatch.setenv('SHPKPR_APPLE_AND_BLACKCURRANT', 'crumble')
    monkeypatch.setenv('SHPKPR_SHPKPR_APPLE_AND_BLACKCURRANT', 'crumble')

    values = load_values_from_environment("SHPKPR_")

    assert 'BANANA' not in values
    assert 'STRAWBERRY' in values
    assert values['STRAWBERRY'] == 'cheesecake'
    assert 'APPLE_AND_BLACKCURRANT' in values
    assert values['APPLE_AND_BLACKCURRANT'] == 'crumble'
    assert 'SHPKPR_APPLE_AND_BLACKCURRANT' in values
    assert values['SHPKPR_APPLE_AND_BLACKCURRANT'] == 'crumble'


def test_render_json_template_valid(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"type_of_muffin": "{{ MUFFIN_TYPE }}"}',
    )

    rendered_template = render_json_template(template_path, template_name, **{"MUFFIN_TYPE": "banana"})
    assert "type_of_muffin" in rendered_template
    assert rendered_template["type_of_muffin"] == "banana"


def test_render_json_template_invalid_json_unquoted_string(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"type_of_muffin": {{ MUFFIN_TYPE }}}',
    )

    with pytest.raises(InvalidJSONError):
        render_json_template(template_path, template_name, **{"MUFFIN_TYPE": "banana"})


def test_render_json_template_invalid_json_missing_value(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"type_of_muffin": {{ MUFFIN_TYPE }}}',
    )

    with pytest.raises(InvalidJSONError):
        render_json_template(template_path, template_name, **{"MUFFIN_TYPE": ""})


def test_render_json_template_missing_value_raises(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"type_of_muffin": "{{ MUFFIN_TYPE }}"}',
    )

    with pytest.raises(UndefinedError):
        render_json_template(template_path, template_name, **{})


def test_render_json_template_all_env(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '''
        {
            "types_of_muffin": {
                {% for k, v in _all_env|filter_items("MUFFIN_", True) %}
                "{{ k.lower() }}": {{ v }}{% if loop.last == False %},{% endif %}
                {% endfor %}
            }
        }
        ''',
    )

    rendered_template = render_json_template(template_path, template_name, **{
        "MUFFIN_BLUEBERRY": 4,
        "MUFFIN_BANANA": 7,
        "MUFFIN_CHOCOLATE": 12,
        "DONUT_STRAWBERRY": 9,
    })
    assert "types_of_muffin" in rendered_template
    assert rendered_template["types_of_muffin"]["blueberry"] == 4
    assert rendered_template["types_of_muffin"]["banana"] == 7
    assert rendered_template["types_of_muffin"]["chocolate"] == 12


def test_render_json_template_require_int(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_int }}}',
    )

    rendered_template = render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "1"})
    assert rendered_template['muffin_count'] == 1


def test_render_json_template_require_int_requires_int(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_int }}}',
    )

    with pytest.raises(IntegerRequired):
        render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "one muffin"})


def test_render_json_template_require_int_min_constraint(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_int(min=50) }}}',
    )

    rendered_template = render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "60"})
    assert rendered_template['muffin_count'] == 60


def test_render_json_template_require_int_min_constraint_raises(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_int(min=50) }}}',
    )

    with pytest.raises(IntegerTooSmall):
        render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "40"})


def test_render_json_template_require_int_max_constraint(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_int(max=50) }}}',
    )

    rendered_template = render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "-60"})
    assert rendered_template['muffin_count'] == -60


def test_render_json_template_require_int_max_constraint_raises(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_int(max=50) }}}',
    )

    with pytest.raises(IntegerTooLarge):
        render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "60"})


def test_render_json_template_require_float(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_float }}}',
    )

    rendered_template = render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "1.01"})
    assert rendered_template['muffin_count'] == 1.01


def test_render_json_template_require_float_requires_float(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_float }}}',
    )

    with pytest.raises(FloatRequired):
        render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "one muffin"})


def test_render_json_template_require_float_min_constraint(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_float(min=50) }}}',
    )

    rendered_template = render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "60"})
    assert rendered_template['muffin_count'] == 60


def test_render_json_template_require_float_min_constraint_raises(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_float(min=50) }}}',
    )

    with pytest.raises(FloatTooSmall):
        render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "40"})


def test_render_json_template_require_float_max_constraint(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_float(max=50) }}}',
    )

    rendered_template = render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "-60"})
    assert rendered_template['muffin_count'] == -60


def test_render_json_template_require_float_max_constraint_raises(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '{"muffin_count": {{ MUFFIN_COUNT|require_float(max=50) }}}',
    )

    with pytest.raises(FloatTooLarge):
        render_json_template(template_path, template_name, **{"MUFFIN_COUNT": "60"})


def test_render_json_template_with_inheritance_valid(tmpdir):
    _write_template_to_disk(
        tmpdir.mkdir('bases'), 'base.json',
        '{"type_of_muffin": "{% block MUFFIN_TYPE_PLACEHOLDER %}{{ MUFFIN_TYPE }}{% endblock %}"}',
    )
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '''
        {% extends "bases/base.json" %}
        {% block MUFFIN_TYPE_PLACEHOLDER %}{{ ALT_MUFFIN_TYPE }}{% endblock %}
        ''',
    )

    rendered_template = render_json_template(template_path, template_name, **{
        "MUFFIN_TYPE": "banana",
        "ALT_MUFFIN_TYPE": "blueberry",
    })
    assert "type_of_muffin" in rendered_template
    assert rendered_template["type_of_muffin"] == "blueberry"


def test_render_json_template_with_inheritance_no_parent(tmpdir):
    template_path, template_name = _write_template_to_disk(
        tmpdir, 'template.json',
        '''
        {% extends "bases/base.json" %}
        {% block MUFFIN_TYPE_PLACEHOLDER %}{{ ALT_MUFFIN_TYPE }}{% endblock %}
        ''',
    )

    with pytest.raises(MissingTemplateError):
        render_json_template(template_path, template_name, **{})



# stdlib imports
from collections import Counter

# local imports
from shpkpr import exceptions


class ValidationError(exceptions.ShpkprException):
    exit_code = 2

    def format_message(self):
        return 'Unable to validate deployment: %s' % self.message


class Validator(object):
    """Validator is used to run pre-flight checks on a pending deployment.

    The validation is executed after the template is rendered but before the
    application definition is prepared/transformed for deployment.
    """

    def __init__(self, marathon_client, validators=None):
        if validators is None:
            validators = [AppDefinitionValidator, MarathonStateValidator]
        self.validators = [v(marathon_client) for v in validators]

    def validate(self, app_definition):
        for validator in self.validators:
            validator.validate(app_definition)


class AppDefinitionValidator(object):
    """Ensure that an app definition has the minimum set of labels required for
    a bluegreen deploy.

    These labels ensure shpkpr is able to manage and track "groups" of otherwise
    independent/standalone applications that have previously been deployed to
    Marathon.
    """

    REQUIRED_LABELS = [
        "HAPROXY_DEPLOYMENT_GROUP"
    ]

    ERROR_MESSAGE = "Missing label(s) from application definition: {0}"

    def __init__(self, *args, **kwargs):
        pass

    def validate(self, app_definition):
        missing_labels = []
        labels = app_definition.get('labels', {})
        for required_label in self.REQUIRED_LABELS:
            if required_label in labels:
                continue
            missing_labels.append(required_label)
        if missing_labels:
            msg = self.ERROR_MESSAGE.format(", ".join(missing_labels))
            raise ValidationError(msg)


class MarathonStateValidator(object):
    """Ensure that the application is in a clean state on Marathon before
    deploying.

    This validator checks that exactly one application is running in the target
    deployment group. If more than one application stack (typically both blue
    and green) are detected then it may indicate that another deploy is in
    progress or that a previous failed but wasn't cleaned up. Either way a
    deployment should not continue under those circumstances.
    """

    ERROR_MESSAGE = (
        "More than one active application stack detected on Marathon, please "
        "resolve before continuing with your deploy. This may indicate that "
        "another deploy is already in progress or that a previous one has "
        "failed."
    )

    def __init__(self, marathon_client, *args, **kwargs):
        self.marathon_client = marathon_client

    def validate(self, app_definition):
        def get_deployment_group(app_definition):
            labels = app_definition.get('labels', {})
            return labels.get('HAPROXY_DEPLOYMENT_GROUP')

        app_group = get_deployment_group(app_definition)

        remote_apps = self.marathon_client.list_applications()
        remote_groups = [get_deployment_group(app) for app in remote_apps]

        if Counter(remote_groups)[app_group] > 1:
            raise ValidationError(self.ERROR_MESSAGE)



# third-party imports
import mock

# local imports
from shpkpr.marathon_lb import resolver


def _mock_gethostbyname_ex(aliaslist=None, ipaddrlist=None):
    """Returns a mocked socket.gethostbyname_ex function for testing use
    """
    if aliaslist is None:
        aliaslist = []
    if ipaddrlist is None:
        ipaddrlist = ['127.0.0.1']
    return lambda hostname: (hostname, aliaslist, ipaddrlist)


@mock.patch('socket.gethostbyname_ex')
def test_resolve(gethostbyname_ex):
    ipaddrlist = ['127.0.0.1', '127.0.0.2']
    gethostbyname_ex.side_effect = _mock_gethostbyname_ex(ipaddrlist=ipaddrlist)

    actual_urls = resolver.resolve('http://foobar.com:1234')
    expected_urls = ['http://127.0.0.1:1234', 'http://127.0.0.2:1234']

    assert actual_urls == expected_urls



# stdlib imports
import functools

# third-party imports
import click
import pytest
from click.testing import CliRunner

# local imports
from shpkpr.cli import options


@pytest.fixture
def runner():
    @click.command()
    @options.marathon_client
    def _test(marathon_client, **kw):
        pass

    runner = CliRunner()
    return functools.partial(runner.invoke, _test)


def test_marathon_client_fails_with_no_args(runner):
    result = runner()
    assert result.exit_code == 2


def test_marathon_client_succeeds_with_marathon_url(runner):
    result = runner(["--marathon-url", "http://example.com"])
    assert result.exit_code == 0


def test_marathon_client_succeeds_with_marathon_url_and_basic_auth(runner):
    result = runner([
        "--marathon-url", "https://example.com",
        "--username", "someuser",
        "--password", "somepassword",
    ])

    assert result.exit_code == 0


def test_marathon_client_fails_with_marathon_url_and_basic_auth_without_ssl(runner):
    result = runner([
        "--marathon-url", "http://example.com",
        "--username", "someuser",
        "--password", "somepassword",
    ])

    assert result.exit_code == 2


def test_marathon_client_prompts_for_basic_auth_password(runner):
    result = runner([
        "--marathon-url", "https://example.com",
        "--username", "someuser",
    ], input="somepassword\n")

    assert result.exit_code == 0


def test_marathon_client_succeeds_fails_when_missing_basic_auth_username(runner):
    result = runner([
        "--marathon-url", "https://example.com",
        "--password", "somepassword",
    ])

    assert result.exit_code == 2



# stdlib imports
import copy
import json

# third-party imports
import pytest


def _check_exits_zero(result):
    assert result.exit_code == 0


def _check_exits_non_zero(result):
    assert result.exit_code != 0


def _check_output_contains(result, match):
    assert match in result.output


def _check_output_does_not_contain(result, match):
    assert match not in result.output


@pytest.mark.integration
def test_help(runner, env):
    result = runner(["--help"], env=env)
    _check_exits_zero(result)
    _check_output_contains(result, "Usage:")


@pytest.mark.integration
def test_show(runner, env):
    result = runner(["apps", "show"], env=env)
    _check_exits_non_zero(result)
    _check_output_contains(result, "Unable to retrieve application details from marathon: does not exist.")


@pytest.mark.integration
def test_deploy_custom_template(runner, env):
    _tmpl_path = "tests/fixtures/templates/marathon/test.json.tmpl"
    result = runner(["apps", "deploy", "-t", _tmpl_path, "RANDOM_LABEL=some_value"], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_show_again(runner, env):
    result = runner(["apps", "show"], env=env)
    _check_exits_zero(result)
    _check_output_contains(result, '"id": "/%s"' % env['SHPKPR_APPLICATION'])


@pytest.mark.integration
def test_deploy_default_template(runner, env):
    cmd = ["apps",
           "deploy",
           "LABEL_RANDOM=some_value",
           "DOCKER_EXPOSED_PORT=8080",
           "MARATHON_HEALTH_CHECK_PATH=/"]
    result = runner(cmd, env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_show_yet_again(runner, env):
    result = runner(["apps", "show", "-a", env['SHPKPR_MARATHON_APP_ID']], env=env)
    _check_exits_zero(result)
    _check_output_contains(result, '"id": "/%s"' % env['SHPKPR_MARATHON_APP_ID'])
    _check_output_contains(result, '"RANDOM": "some_value"')


@pytest.mark.integration
def test_run(runner, env):
    result = runner(["apps", "run", "env", "ENV_TESTY_MC_TESTFACE=testy"], env=env)
    _check_exits_zero(result)
    _check_output_contains(result, "TESTY_MC_TESTFACE=testy")


@pytest.mark.integration
def test_cron_show(runner, env):
    result = runner(["cron", "show"], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_set(runner, env):
    _tmpl_path = 'tests/fixtures/templates/chronos/test-chronos.json.tmpl'
    result = runner(["cron", "set", "--template", _tmpl_path, "CHRONOS_JOB_NAME=shpkpr-test-job"], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_show_after_add(runner, env):
    result = runner(["cron", "show"], env=env)

    _check_output_contains(result, '"name": "shpkpr-test-job"')
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_run(runner, env):
    result = runner(["cron", "run", "shpkpr-test-job"], env=env)

    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_delete_tasks(runner, env):
    result = runner(["cron", "delete-tasks", "shpkpr-test-job"], env=env)

    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_delete(runner, env):
    result = runner(["cron", "delete", "shpkpr-test-job"], env=env)

    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_show_after_delete(runner, env):
    result = runner(["cron", "show"], env=env)

    _check_output_does_not_contain(result, '"name": "shpkpr-test-job"')
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_set_default_template(runner, env):
    cmd = ["cron",
           "set",
           "CHRONOS_JOB_NAME=shpkpr-test-job",
           "CHRONOS_OWNER=example@example.com",
           'CHRONOS_CMD=echo \\"Hello World!\\"']
    result = runner(cmd, env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_show_default_template_after_add(runner, env):
    result = runner(["cron", "show"], env=env)

    _check_output_contains(result, '"name": "shpkpr-test-job"')
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_delete_default_template(runner, env):
    result = runner(["cron", "delete", "shpkpr-test-job"], env=env)

    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_show_default_template_after_delete(runner, env):
    result = runner(["cron", "show"], env=env)

    _check_output_does_not_contain(result, '"name": "shpkpr-test-job"')
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_set_with_secrets(runner, env):
    _tmpl_path = 'tests/fixtures/templates/chronos/test-chronos-secrets.json.tmpl'
    result = runner(["cron", "set", "--template", _tmpl_path, "CHRONOS_JOB_NAME=shpkpr-test-job"], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_show_secrets_after_add(runner, env):
    result = runner(["cron", "show", "-j", "shpkpr-test-job"], env=env)

    env_vars = json.loads(result.output)[0].get("environmentVariables", [])
    assert {"name": "MY_SECRET", "value": "bar"} in env_vars
    _check_exits_zero(result)


@pytest.mark.integration
def test_cron_delete_secrets(runner, env):
    result = runner(["cron", "delete", "shpkpr-test-job"], env=env)

    _check_exits_zero(result)


@pytest.mark.integration
def test_bluegreen_deploy(runner, env):
    _tmpl_path = "tests/fixtures/templates/marathon/test-bluegreen.json.tmpl"
    result = runner(["apps", "deploy", "--strategy", "bluegreen", "-t", _tmpl_path], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_ensure_bluegreen_deployed(runner, env):
    local_env = copy.deepcopy(env)
    local_env["SHPKPR_APPLICATION"] = "shpkpr-test/integration-test-bluegreen-blue"

    result = runner(["apps", "show"], env=local_env)
    _check_exits_zero(result)
    _check_output_contains(result, "shpkpr-test/integration-test-bluegreen-blue")


@pytest.mark.integration
def test_bluegreen_deploy_with_existing_app(runner, env):
    _tmpl_path = "tests/fixtures/templates/marathon/test-bluegreen.json.tmpl"
    result = runner(["apps", "deploy", "--strategy", "bluegreen", "-t", _tmpl_path], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_ensure_bluegreen_swapped(runner, env):
    local_env = copy.deepcopy(env)
    local_env["SHPKPR_APPLICATION"] = "shpkpr-test/integration-test-bluegreen-green"

    result = runner(["apps", "show"], env=local_env)
    _check_exits_zero(result)
    _check_output_contains(result, "shpkpr-test/integration-test-bluegreen-green")


@pytest.mark.integration
def test_bluegreen_deploy_default_template(runner, env):
    cmd = ["apps",
           "deploy",
           "--strategy",
           "bluegreen",
           "DOCKER_EXPOSED_PORT=8080",
           "MARATHON_APP_ID=shpkpr-test/integration-test-bluegreen-default-template",
           "MARATHON_HEALTH_CHECK_PATH=/",
           "MARATHON_SERVICE_PORT=11092",
           "MARATHON_SERVICE_PORT_ALT=11093",
           "HAPROXY_VHOST=shpkpr-test-bgdt.somedomain.com"]
    result = runner(cmd, env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_ensure_bluegreen_default_template_deployed(runner, env):
    local_env = copy.deepcopy(env)
    local_env["SHPKPR_APPLICATION"] = "shpkpr-test/integration-test-bluegreen-default-template-blue"

    result = runner(["apps", "show"], env=local_env)
    _check_exits_zero(result)
    _check_output_contains(result, local_env["SHPKPR_APPLICATION"])


@pytest.mark.integration
def test_bluegreen_deploy_with_existing_app_default_template(runner, env):
    cmd = ["apps",
           "deploy",
           "--strategy",
           "bluegreen",
           "DOCKER_EXPOSED_PORT=8080",
           "MARATHON_APP_ID=shpkpr-test/integration-test-bluegreen-default-template",
           "MARATHON_HEALTH_CHECK_PATH=/",
           "MARATHON_SERVICE_PORT=11092",
           "MARATHON_SERVICE_PORT_ALT=11093",
           "HAPROXY_VHOST=shpkpr-test-bgdt.somedomain.com"]
    result = runner(cmd, env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_ensure_bluegreen_default_template_swapped(runner, env):
    local_env = copy.deepcopy(env)
    local_env["SHPKPR_APPLICATION"] = "shpkpr-test/integration-test-bluegreen-default-template-green"

    result = runner(["apps", "show"], env=local_env)
    _check_exits_zero(result)
    _check_output_contains(result, local_env["SHPKPR_APPLICATION"])


@pytest.mark.integration
def test_bluegreen_autoports_deploy(runner, env):
    _tmpl_path = "tests/fixtures/templates/marathon/test-bluegreen-autoports.json.tmpl"
    result = runner(["apps", "deploy", "--strategy", "bluegreen", "-t", _tmpl_path], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_ensure_bluegreen_autoports_deploy(runner, env):
    local_env = copy.deepcopy(env)
    local_env["SHPKPR_APPLICATION"] = "shpkpr-test/integration-test-bluegreen-blue"

    result = runner(["apps", "show"], env=local_env)
    _check_exits_zero(result)
    _check_output_contains(result, "shpkpr-test/integration-test-bluegreen-blue")


@pytest.mark.integration
def test_bluegreen_autoports_deploy_with_existing_app(runner, env):
    _tmpl_path = "tests/fixtures/templates/marathon/test-bluegreen-autoports.json.tmpl"
    result = runner(["apps", "deploy", "--strategy", "bluegreen", "-t", _tmpl_path], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_ensure_bluegreen_autoports_swapped(runner, env):
    local_env = copy.deepcopy(env)
    local_env["SHPKPR_APPLICATION"] = "shpkpr-test/integration-test-bluegreen-green"

    result = runner(["apps", "show"], env=local_env)
    _check_exits_zero(result)
    _check_output_contains(result, "shpkpr-test/integration-test-bluegreen-green")


@pytest.mark.integration
def test_bluegreen_autoports_deploy_reset(runner, env):
    _tmpl_path = "tests/fixtures/templates/marathon/test-bluegreen-autoports.json.tmpl"
    result = runner(["apps", "deploy", "--strategy", "bluegreen", "-t", _tmpl_path], env=env)
    _check_exits_zero(result)


@pytest.mark.integration
def test_ensure_bluegreen_autoports_reset(runner, env):
    local_env = copy.deepcopy(env)
    local_env["SHPKPR_APPLICATION"] = "shpkpr-test/integration-test-bluegreen-blue"

    result = runner(["apps", "show"], env=local_env)
    _check_exits_zero(result)
    _check_output_contains(result, "shpkpr-test/integration-test-bluegreen-blue")



# third-party imports
import pytest

# local imports
from shpkpr.deployment.bluegreen.prepare import prepare_app_definition


@pytest.fixture
def app_definition(json_fixture):
    """Prepare an app definition for deployment.

    This fixture prepares an app and assumes that an existing stack is present
    on Marathon.
    """
    new_app_definition_fixture = json_fixture("marathon/bluegreen_app_new")
    old_app_definition_fixture = json_fixture("marathon/bluegreen_app_existing")
    app_states_fixture = json_fixture("valid_apps")
    info_definition_fixture = json_fixture("marathon/info")
    return prepare_app_definition(new_app_definition_fixture,
                                  old_app_definition_fixture,
                                  app_states_fixture['apps'],
                                  info_definition_fixture)


def test_color(app_definition):
    """Test that the app's color is set as a label
    """
    actual = app_definition['labels']['HAPROXY_DEPLOYMENT_COLOUR']
    expected = "green"

    assert actual == expected


def test_id(app_definition):
    """Test that the app's ID is correctly transformed
    """
    actual = app_definition['id']
    expected = "my-app-green"

    assert actual == expected


def test_docker_service_port(app_definition):
    """Test that the service port is set correctly when using docker
    """
    actual = app_definition['container']['docker']['portMappings'][0]['servicePort']
    expected = 11091

    assert actual == expected


def test_labels_id(app_definition):
    """Test that the app's original ID is preserved as a label
    """
    actual = app_definition['labels']['HAPROXY_APP_ID']
    expected = "my-app"

    assert actual == expected


def test_labels_port(app_definition):
    """Test that the app's original port is preserved as a label
    """
    actual = app_definition['labels']['HAPROXY_0_PORT']
    expected = "11090"

    assert actual == expected


def test_labels_alt_port(app_definition):
    """Test that the app's original port is preserved as a label
    """
    actual = app_definition['labels']['HAPROXY_DEPLOYMENT_ALT_PORT']
    expected = "11091"

    assert actual == expected



# stdlib imports
import json
import os

# third-party imports
import mock
import pytest
import responses

# local imports
from shpkpr.marathon import ClientError
from shpkpr.marathon import DeploymentNotFound
from shpkpr.marathon import DryRun
from shpkpr.marathon import MarathonClient


def _load_json_fixture(name):
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir, "fixtures", name + ".json")
    with open(path, 'r') as f:
        fixture = json.loads(f.read())
    return fixture


@responses.activate
def test_get_application():
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=200,
                  json=_load_json_fixture("valid_app"))

    client = MarathonClient("http://marathon.somedomain.com:8080")
    application = client.get_application('test-app')
    assert application['id'] == "/test-app"
    assert application['cmd'] is None


@responses.activate
def test_get_application_not_found():
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=404,
                  json={"message": "App '/test-app' does not exist"})

    client = MarathonClient("http://marathon.somedomain.com:8080")
    with pytest.raises(ClientError):
        client.get_application('test-app')


@responses.activate
def test_get_application_internal_server_error():
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=500,
                  body="Internal Server Error")

    client = MarathonClient("http://marathon.somedomain.com:8080")
    with pytest.raises(ClientError):
        client.get_application('test-app')


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.check')
def test_deploy(mock_deployment_check):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=201,
                  json=_load_json_fixture("deployment"))
    mock_deployment_check.return_value = True

    client = MarathonClient("http://marathon.somedomain.com:8080")
    deployment = client.deploy({"id": "test-app"})
    assert deployment.wait(check_interval_secs=0.1)


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.check')
def test_deploy_multiple(mock_deployment_check):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/',
                  status=201,
                  json=_load_json_fixture("deployment"))
    mock_deployment_check.return_value = True

    client = MarathonClient("http://marathon.somedomain.com:8080")
    deployment = client.deploy([{"id": "test-app"}, {"id": "test-app-2"}])
    assert deployment.wait(check_interval_secs=0.1)


@responses.activate
@mock.patch('shpkpr.marathon.MarathonDeployment.check')
def test_deploy_single_app_list(mock_deployment_check):
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=201,
                  json=_load_json_fixture("deployment"))
    mock_deployment_check.return_value = True

    client = MarathonClient("http://marathon.somedomain.com:8080")
    deployment = client.deploy([{"id": "test-app"}])
    assert deployment.wait(check_interval_secs=0.1)


@responses.activate
def test_deploy_conflict():
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=409,
                  json=_load_json_fixture("deployment_in_progress"))

    client = MarathonClient("http://marathon.somedomain.com:8080")
    with pytest.raises(ClientError):
        client.deploy({"id": "test-app"})


@responses.activate
def test_deploy_dry_run():
    client = MarathonClient("http://marathon.somedomain.com:8080", dry_run=True)
    with pytest.raises(DryRun):
        client.deploy({"id": "test-app"})


@responses.activate
def test_deploy_unprocessable_entity():
    responses.add(responses.PUT,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=422,
                  json=_load_json_fixture("validation_errors"))

    client = MarathonClient("http://marathon.somedomain.com:8080")
    with pytest.raises(ClientError):
        client.deploy({"id": "test-app"})


@responses.activate
def test_get_deployment():
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/deployments',
                  status=200,
                  json=_load_json_fixture("deployments"))

    client = MarathonClient("http://marathon.somedomain.com:8080")
    deployment = client.get_deployment("97c136bf-5a28-4821-9d94-480d9fbb01c8")
    assert deployment is not None
    assert deployment['id'] == "97c136bf-5a28-4821-9d94-480d9fbb01c8"


@responses.activate
def test_get_deployment_not_found():
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/deployments',
                  status=200,
                  json=_load_json_fixture("deployments"))

    client = MarathonClient("http://marathon.somedomain.com:8080")
    with pytest.raises(DeploymentNotFound):
        client.get_deployment("1234")


@responses.activate
def test_basic_auth_headers_present():
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=200,
                  json=_load_json_fixture("valid_app"))

    client = MarathonClient("http://marathon.somedomain.com:8080",
                            "some-username",
                            "some-password")
    client.get_application('test-app')

    assert "Authorization" in responses.calls[0].request.headers


@responses.activate
def test_basic_auth_headers_not_present():
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=200,
                  json=_load_json_fixture("valid_app"))

    client = MarathonClient("http://marathon.somedomain.com:8080")
    client.get_application('test-app')

    assert "Authorization" not in responses.calls[0].request.headers



# stdlib imports
from setuptools import find_packages
from setuptools import setup


try:
    with open('README.rst') as readme_file:
        readme = readme_file.read()
except IOError:
    # README is not always available at the time this file is imported. This
    # really only applies during an initial container build when we copy in
    # setup.py but not the README. This is mostly to help with caching and
    # keeping the time required to run the tests to a minimum. This should never
    # make it into an actual build or release package.
    readme = "Not Available"


setup(
    name='shpkpr',
    version='4.1.0',
    description='shpkpr is a command-line tool designed to manage applications running on Marathon',
    long_description=readme,
    author='ShopKeep.com Inc.',
    author_email='developers@shopkeep.com',
    url='https://github.com/shopkeep/shpkpr',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'PyYAML>=3.10.0',
        'cached-property>=1.3.0, <2',
        'chronos-python>=1.1.0',
        'click>=6.0, <7',
        'docker>=3.0.0',
        'hvac>=0.2.17',
        'jinja2>=2.6, <3',
        'jsonschema>=2.6, <3',
        'python-slugify>=1.2.4',
        'six>=1.7.0',
        'requests>=2.4.2',
    ],
    license='MIT',
    zip_safe=False,
    keywords='shpkpr mesos marathon chronos',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    entry_points='''
        [console_scripts]
        shpkpr=shpkpr.cli.entrypoint:cli
    ''',
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=[
        'freezegun',
        'mock',
        'pytest',
        'responses',
    ],
)



# local imports
from shpkpr import exceptions

# third-party imports
import slugify as libslugify


def filter_items(value, startswith=None, strip_prefix=False):
    """Jinja2 filter used to filter a dictionary's keys by specifying a
    required prefix.

    Returns a list of key/value tuples.

    .. code-block:: jinja

        {{ my_dict|filter_items }}
        {{ my_dict|filter_items("MY_PREFIX_") }}
        {{ my_dict|filter_items("MY_PREFIX_", True) }}

    This is most useful in combination with the special
    :ref:`_all_env <all_env>` variable that shpkpr injects into every template.
    For example, to iterate over only the template variables that start with
    ``LABEL_`` you could do:

    .. code-block:: jinja

        {% for k, v in _all_env|filter_items("LABEL_", strip_prefix=True) %}
            "{{k}}": "{{v}}",
        {% endfor %}

    """
    if startswith is not None:
        value = [x for x in value.items() if x[0].startswith(startswith)]
    else:
        value = value.items()

    if startswith is not None and strip_prefix:
        value = [(x[0].replace(startswith, "", 1), x[1]) for x in value]

    return value


class IntegerRequired(exceptions.ShpkprException):
    exit_code = 2


class IntegerTooSmall(exceptions.ShpkprException):
    exit_code = 2


class IntegerTooLarge(exceptions.ShpkprException):
    exit_code = 2


@exceptions.rewrap(ValueError, IntegerRequired)
def require_int(value, min=None, max=None):
    """Jinja2 filter used to enforce an integer type in a template.

    Accepts optional min/max constraints.

    .. code-block:: jinja

        {{ my_integer_value|require_int }}
        {{ my_integer_value|require_int(min=0) }}
        {{ my_integer_value|require_int(min=0, max=20) }}
        {{ my_integer_value|require_int(max=20) }}
    """
    value = int(value)

    # check the integer falls within the allowed range if set
    if min is not None and value < min:
        raise IntegerTooSmall("%d is smaller than the minimum allowed value of %d" % (value, min))
    if max is not None and value > max:
        raise IntegerTooLarge("%d is larger than the maximum allowed value of %d" % (value, max))

    return value


class FloatRequired(exceptions.ShpkprException):
    exit_code = 2


class FloatTooSmall(exceptions.ShpkprException):
    exit_code = 2


class FloatTooLarge(exceptions.ShpkprException):
    exit_code = 2


@exceptions.rewrap(ValueError, FloatRequired)
def require_float(value, min=None, max=None):
    """Jinja2 filter used to enforce an float type in a template.

    Accepts optional min/max constraints.

    .. code-block:: jinja

        {{ my_float_value|require_float }}
        {{ my_float_value|require_float(min=0) }}
        {{ my_float_value|require_float(min=0, max=20) }}
        {{ my_float_value|require_float(max=20) }}
    """
    value = float(value)

    # check the float falls within the allowed range if set
    if min is not None and value < min:
        raise FloatTooSmall("%f is smaller than the minimum allowed value of %f" % (value, min))
    if max is not None and value > max:
        raise FloatTooLarge("%f is larger than the maximum allowed value of %f" % (value, max))

    return value


def slugify(value, *args, **kwargs):
    """Jinja2 filter used to slugify a string.

    All arguments are passed directly to python-slugify_'s ``slugify``
    function, see the library's documentation_ for further information.

    .. code-block:: jinja

        {{ my_string|slugify }}
        {{ my_string|slugify(max_length=20) }}

    .. _python-slugify: https://pypi.python.org/pypi/python-slugify
    .. _documentation: https://github.com/un33k/python-slugify#how-to-use
    """
    return libslugify.slugify(value, *args, **kwargs)



"""Option parameter definitions for shpkpr.

All option parameters should be defined in this module, commands should not
define their options directly, instead all options should be defined here and
imported as required. This helps maintain consistency where a single option is
used for multiple commands.
"""
# stdlib imports
import os
from distutils.version import StrictVersion

# third-party imports
import click
import hvac
from chronos import ChronosClient
from six.moves.urllib import parse

# local imports
from shpkpr.cli.decorators import multioption
from shpkpr.cli.entrypoint import CONTEXT_SETTINGS
from shpkpr.cli.formatter import OutputFormatter
from shpkpr.marathon import MarathonClient
from shpkpr.marathon_lb import MarathonLBClient


application_id = click.option(
    '-a', '--application',
    'application_id',
    envvar="{0}_APPLICATION".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    help=(
        "ID of the Marathon application to target. This should be the full "
        "path of the application (including slashes) e.g. ``/mygroup/myapp``. "
        "The first/root slash is optional, shpkpr will add it if missing."
    ),
)


dry_run = click.option(
    '--dry-run',
    is_flag=True,
    help=(
        "Enables \"dry run\" mode. When enabled, shpkpr will avoid making "
        "changes to any remote systems. This can be used to safely test "
        "deployments or jobs before performing any destructive operations. "
        "Exactly what constitues a \"dry run\" varies by operation type and "
        "deployment strategy, see command documentation for further "
        "information."
    ),
)


env_prefix = click.option(
    '-e', '--env-prefix',
    'env_prefix',
    default=CONTEXT_SETTINGS['auto_envvar_prefix'],
    show_default=True,
    help="Prefix used to filter environment variables used for templating.",
)


force = click.option(
    '--force',
    is_flag=True,
    help='Force update even if a deployment is in progress.',
)


deployment_strategy = click.option(
    '--strategy',
    'deployment_strategy',
    type=click.Choice(["standard", "bluegreen"]),
    help='Deployment strategy to utilise.',
    envvar="{0}_DEPLOYMENT_STRATEGY".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    default="standard",
    show_default=True,
)


template_path = click.option(
    '--template-dir',
    'template_path',
    envvar="{0}_TEMPLATE_DIR".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    type=str,
    default=os.getcwd(),
    help="Base directory in which your templates are stored.",
)


template_names = click.option(
    '-t', '--template',
    'template_names',
    envvar="{0}_TEMPLATES".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    type=str,
    help="Path of the template to use for deployment.",
    multiple=True,
)

chronos_url = click.option(
    '--chronos-url',
    'chronos_url',
    envvar="{0}_CHRONOS_URL".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    required=True,
    help='URL of the Chronos endpoint to use',
    callback=lambda c, p, v: parse.urlparse(v, scheme="https"),
)


def _validate_chronos_version(ctx, param, value):
    """chronos-python requires a version "prefix" to be used, so we convert the
    target Chronos version to the appropriate prefix.
    """
    if StrictVersion(value) >= StrictVersion('3.0.0'):
        return 'v1'
    return None


chronos_version = click.option(
    '--chronos-version',
    'chronos_version',
    envvar="{0}_CHRONOS_VERSION".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    required=False,
    help='Verson of the Chronos endpoint to use',
    default='3.0.2',
    show_default=True,
    callback=_validate_chronos_version,
)

job_name = click.option(
    '-j', '--job-name',
    type=str,
    help='Restrict command to specific job.',
)


marathon_lb_client = click.option(
    '--marathon-lb-url',
    'marathon_lb_client',
    envvar="{0}_MARATHON_LB_URL".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    default=None,
    help="URL for Marathon-LB used during blue/green deployment.",
    callback=lambda c, p, v: None if v is None else MarathonLBClient(v)
)


timeout = click.option(
    '--timeout',
    'timeout',
    type=int,
    help='Maximum amount of time (in seconds) to wait for a deployment to finish before aborting.',
    envvar="{0}_TIMEOUT".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    default=300,
    show_default=True,
)

output_formatter = click.option(
    '--output-format',
    'output_formatter',
    type=click.Choice(["json", "yaml"]),
    help='Serialization format to use when printing config data to stdout.',
    envvar="{0}_OUTPUT_FORMAT".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    default="json",
    show_default=True,
    callback=lambda c, p, v: OutputFormatter(v)
)


def _validate_authentication(ctx, url, service_name):
    """Validates that authentication-related options required to initialise
    marathon/chronos clients have been set properly and securely.
    """
    allow_insecure_auth = ctx.params["allow_insecure_auth"]
    username = ctx.params["username"]
    password = ctx.params["password"]

    insecure_configuration = all([
        bool(username) or bool(password),
        not url.startswith("https://"),
        not allow_insecure_auth,
    ])

    if insecure_configuration:
        raise click.UsageError(
            "HTTPS is strongly recommended when using basic authentication (as "
            "credentials are sent unencrypted over the network).\n\nIf you "
            "want to allow insecure communications regardless, you can use the "
            "command-line flag: `--allow-insecure-auth`"
        )

    if username is not None and password is None:
        password = click.prompt("Please enter your {0} password".format(service_name))
    if password is not None and username is None:
        raise click.UsageError("A username is required to use HTTP Basic Authentication")


username = click.option(
    '--username',
    'username',
    type=str,
    help='Username used for HTTP Basic Authentication to Marathon/Chronos.',
    envvar="{0}_USERNAME".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    default=None,
)

password = click.option(
    '--password',
    'password',
    type=str,
    help='Password used for HTTP Basic Authentication to Marathon/Chronos.',
    envvar="{0}_PASSWORD".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    default=None,
)

allow_insecure_auth = click.option(
    '--allow-insecure-auth',
    is_flag=True,
    help='Allow HTTP basic authentication when not using SSL.',
)

marathon_url = click.option(
    '--marathon-url',
    'marathon_url',
    envvar="{0}_MARATHON_URL".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    required=True,
    help="URL of the Marathon API to use.",
)


def _validate_marathon_client(ctx, _, __):
    """Validates that all options required to initialise a marathon client have
    been set properly and securely.

    A client is then initialised and returned to the command function.
    """
    _c = ctx.params
    _validate_authentication(ctx, _c["marathon_url"], "Marathon")
    return MarathonClient(_c["marathon_url"],
                          _c["username"],
                          _c["password"],
                          _c["dry_run"])


marathon_client = multioption(
    name='marathon_client',
    callback=_validate_marathon_client,
    options=[
        dry_run,
        allow_insecure_auth,
        username,
        password,
        marathon_url,
    ],
)


def _validate_chronos_client(ctx, _, __):
    """Validates that all options required to initialise a chronos client have
    been set properly and securely.

    A client is then initialised and returned to the command function.
    """
    _c = ctx.params
    _validate_authentication(ctx, parse.urlunparse(_c["chronos_url"]), "Chronos")
    return ChronosClient(_c["chronos_url"].netloc,
                         proto=_c["chronos_url"].scheme,
                         scheduler_api_version=_c["chronos_version"],
                         username=_c["username"],
                         password=_c["password"])


chronos_client = multioption(
    name='chronos_client',
    callback=_validate_chronos_client,
    options=[
        allow_insecure_auth,
        username,
        password,
        chronos_url,
        chronos_version,
    ],
)


vault_addr = click.option(
    '--vault-addr',
    'vault_addr',
    envvar="{0}_VAULT_ADDR".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    help="URL of the Vault API to use.",
)

vault_token = click.option(
    '--vault-token',
    'vault_token',
    envvar="{0}_VAULT_TOKEN".format(CONTEXT_SETTINGS['auto_envvar_prefix']),
    help="Token used to authenticate with Vault.",
)


def _validate_vault_client(ctx, _, __):
    """Validates that all options required to initialise a vault client have
    been set properly and securely.

    A client is then initialised and returned to the command function.
    """
    return hvac.Client(url=ctx.params['vault_addr'],
                       token=ctx.params['vault_token'])


vault_client = multioption(
    name='vault_client',
    callback=_validate_vault_client,
    options=[
        vault_addr,
        vault_token,
    ],
)



# stdlib imports
import functools
import os

# third-party imports
import pytest
from click.testing import CliRunner

# local imports
from shpkpr.cli.entrypoint import cli


@pytest.fixture(scope="session")
def env():
    # read the required environment variables into a dictionary and assert
    # that they're set appropriately
    env = {
        "SHPKPR_MARATHON_APP_ID": os.environ.get("SHPKPR_MARATHON_APP_ID", None),
        "SHPKPR_MARATHON_URL": os.environ.get("SHPKPR_MARATHON_URL", None),
        "SHPKPR_MARATHON_LB_URL": os.environ.get("SHPKPR_MARATHON_LB_URL", None),
        "SHPKPR_APPLICATION": os.environ.get("SHPKPR_APPLICATION", None),
        "SHPKPR_DEPLOYMENT_GROUP": os.environ.get("SHPKPR_DEPLOYMENT_GROUP", None),
        "SHPKPR_DOCKER_REPOTAG": os.environ.get("SHPKPR_DOCKER_REPOTAG", None),
        "SHPKPR_DOCKER_EXPOSED_PORT": os.environ.get("SHPKPR_DOCKER_EXPOSED_PORT", None),
        "SHPKPR_DEPLOY_DOMAIN": os.environ.get("SHPKPR_DEPLOY_DOMAIN", None),
        "SHPKPR_CHRONOS_URL": os.environ.get("SHPKPR_CHRONOS_URL", None),
        "SHPKPR_CHRONOS_VERSION": os.environ.get("SHPKPR_CHRONOS_VERSION", None),
        "SHPKPR_VAULT_ADDR": os.environ.get("SHPKPR_VAULT_ADDR", None),
        "SHPKPR_VAULT_TOKEN": os.environ.get("SHPKPR_VAULT_TOKEN", None),
    }
    assert None not in env.values()
    return env


@pytest.fixture
def runner():
    runner = CliRunner()
    return functools.partial(runner.invoke, cli)



"""Templating related utilities
"""
# stdlib imports
import json
import os

# third-party imports
import jinja2

# local imports
from shpkpr import exceptions
from shpkpr import template_filters


class InvalidJSONError(exceptions.ShpkprException):
    """Raised when a template can be rendered successfully but does not parse
    as valid JSON afterwards.
    """
    exit_code = 2

    def format_message(self):
        return 'Unable to parse rendered template as JSON, check variables'


class UndefinedError(exceptions.ShpkprException):
    """Raised when a template contains a placeholder for a variable that
    wasn't included in the context dictionary passed in at render time.
    """
    exit_code = 2

    def format_message(self):
        return 'Unable to render template: %s' % self.message


class MissingTemplateError(exceptions.ShpkprException):
    """Raised when a template cannot be loaded for any reason.
    """
    exit_code = 2

    def format_message(self):
        return 'Unable to load template from disk: %s' % self.message


def load_values_from_environment(prefix="", overrides=None):
    """Reads values from the environment.

    If ``prefix`` is a non-empty string, only environment variables with the
    given prefix will be returned. The prefix, if given, will be stripped from
    any returned keys.

    If ``overrides`` is a dict-like object, the key/value pairs it contains
    will be added to the returned dictionary. Any values specified by
    overrides will take precedence over values pulled from the environment
    where the key names clash.
    """
    values = {}

    # add a trailing underscore to the prefix if there isn't one
    prefix = prefix + "_" if prefix and not prefix.endswith("_") else prefix

    # load values from the environment
    for k, v in os.environ.items():
        if k.startswith(prefix):
            values[k.replace(prefix, "", 1)] = v

    # add override values if any passed in
    try:
        for k, v in overrides.items():
            values[k] = v
    except AttributeError:
        pass

    return values


@exceptions.rewrap(ValueError, InvalidJSONError)
@exceptions.rewrap(jinja2.UndefinedError, UndefinedError)
@exceptions.rewrap(jinja2.TemplateNotFound, MissingTemplateError)
def render_json_template(template_path, template_name, **values):
    """Initialise a jinja2 template and render it with the passed-in values.

    The template, once rendered is treated as JSON and converted into a python
    dictionary. If the template is not valid JSON after rendering then an
    exception will be raised.

    If a template defines a placeholder for a variable that is not included in
    `values` an `UndefinedError` will be raised.

    ``template_path`` should be the base directory in which your templates are
    stored.
    ``template_name`` is the name (path) of the template being used to render.
    ``values`` should be regular keyword arguments to the function which will
    be passed to the template at render time.
    """
    # shpkpr ships with a number of built-in templates for each deployment type,
    # so we need to tell jinja where to look for them
    here = os.path.dirname(os.path.abspath(__file__))
    built_in_template_path = os.path.join(here, "resources", "templates")

    # build a new Jinja2 environment so we can inject some custom filters into
    # the template we're rendering.
    template_env = jinja2.Environment(
        undefined=jinja2.StrictUndefined,
        loader=jinja2.FileSystemLoader([
            built_in_template_path,
            template_path,
        ]),
    )
    template_env.filters['filter_items'] = template_filters.filter_items
    template_env.filters['require_int'] = template_filters.require_int
    template_env.filters['require_float'] = template_filters.require_float
    template_env.filters['slugify'] = template_filters.slugify

    template = template_env.get_template(template_name)
    rendered_template = template.render(_all_env=values, **values)
    return json.loads(rendered_template)



# stdlib imports
import json

# third-party imports
import responses
import yaml
from click.testing import CliRunner

# local imports
from shpkpr.cli.entrypoint import cli


def test_no_args():
    runner = CliRunner()
    result = runner.invoke(cli, ['apps', 'show'])

    assert result.exit_code == 2
    assert 'Usage:' in result.output


def test_help():
    runner = CliRunner()
    result = runner.invoke(cli, ['apps', 'show', '--help'])

    assert result.exit_code == 0
    assert 'Usage:' in result.output
    assert 'Show detailed information for one or more applications.' in result.output


@responses.activate
def test_show_single_app(runner, json_fixture):
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=200,
                  json=json_fixture("valid_app"))

    result = runner(['apps', 'show'], env={
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
    })
    assert result.exit_code == 0

    parsed_payload = json.loads(result.output)
    assert isinstance(parsed_payload, dict)
    assert parsed_payload["id"] == "/test-app"


@responses.activate
def test_show_multiple_apps(runner, json_fixture):
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/apps',
                  status=200,
                  json=json_fixture("valid_apps"))

    result = runner(['apps', 'show'], env={
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
    })
    assert result.exit_code == 0

    parsed_payload = json.loads(result.output)
    assert isinstance(parsed_payload, list)

    sorted_app_ids = sorted([x.get('id') for x in parsed_payload])
    assert sorted_app_ids == ["/test-app-a", "/test-app-b", "/test-app-c"]


@responses.activate
def test_show_format_json(runner, json_fixture):
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=200,
                  json=json_fixture("valid_app"))

    result = runner(['apps', 'show'], env={
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
    })
    assert result.exit_code == 0

    parsed_payload = json.loads(result.output)
    assert isinstance(parsed_payload, dict)


@responses.activate
def test_show_format_yaml(runner, json_fixture):
    responses.add(responses.GET,
                  'http://marathon.somedomain.com:8080/v2/apps/test-app',
                  status=200,
                  json=json_fixture("valid_app"))

    result = runner(['apps', 'show'], env={
        'SHPKPR_MARATHON_URL': "http://marathon.somedomain.com:8080",
        'SHPKPR_APPLICATION': 'test-app',
    })
    assert result.exit_code == 0

    parsed_payload = yaml.load(result.output)
    assert isinstance(parsed_payload, dict)



# stdlib imports
import logging
import time

# local imports
from . import stats
from shpkpr import exceptions


logger = logging.getLogger(__name__)


class SwapApplicationTimeout(exceptions.ShpkprException):
    pass


class Waiter(object):
    """Wait for a bluegreen deployment to complete by inspecting Marathon-LB's
    stats endpoint.

    This process ensures that the new application is up, stable, and serving
    traffic, and also that the old application is no longer being routed to at
    the load-balancer layer.
    """

    # amount of time (in seconds) between subsequent polls of the load balancer
    # whilst waiting for traffic to cut over to the new application.
    MARATHON_LB_POLL_INTERVAL = 5

    def __init__(self, marathon_client, marathon_lb_client, new_app_id, old_app_id):
        self.marathon_client = marathon_client
        self.marathon_lb_client = marathon_lb_client
        self.new_app_id = new_app_id
        self.old_app_id = old_app_id

    def check(self):
        """Check if the cutover is complete
        """
        if self.marathon_lb_client.is_reloading():
            logger.info("Waiting for Marathon-LB to settle (reloading detected)")
            return False

        new_app = self.marathon_client.get_application(self.new_app_id)
        old_app = self.marathon_client.get_application(self.old_app_id)
        app_stats = self._fetch_application_stats(new_app)
        if not self._new_app_is_up(app_stats, new_app):
            logger.info("Waiting for new application to come up")
            return False
        if not self._old_app_is_drained(app_stats, old_app):
            logger.info("Waiting for traffic to drain from old application")
            return False
        return True

    def wait(self, deadline):
        """Repeatedly check if all incoming traffic has been cut over to the new
        application.

        Blocks until success or ``deadline`` is reached.
        """
        _msg = "Waiting for traffic to cut over from `{0}` to `{1}`"
        logger.info(_msg.format(self.old_app_id, self.new_app_id))

        while not self.check():
            if time.time() >= deadline:
                raise SwapApplicationTimeout('Max wait Time Exceeded')
            time.sleep(self.MARATHON_LB_POLL_INTERVAL)

        logger.info("Traffic successfully routed to new application")

    def _fetch_application_stats(self, app_definition):
        """Fetch stats from HAProxy for the current application.
        """
        all_stats = self.marathon_lb_client.fetch_stats()
        return stats.ApplicationStats(app_definition, all_stats)

    def _new_app_is_up(self, app_stats, app):
        """Check that all tasks for the new application are up, are stable, and
        are serving live traffic.
        """
        haproxy_count = self.marathon_lb_client.instance_count
        checker = ListenerCheck(app_stats, app, haproxy_count)
        return checker.check("UP")

    def _old_app_is_drained(self, app_stats, app):
        """Check that all tasks for the old application have been drained and
        are no longer serving live traffic.
        """
        haproxy_count = self.marathon_lb_client.instance_count
        checker = ListenerCheck(app_stats, app, haproxy_count)
        return checker.check("MAINT", predicate=_has_no_pending_requests)


def _has_no_pending_requests(listener):
    """Returns `True` if the given listener has zero currently queued requests
    and zero currently active sessions, `False` otherwise.
    """
    return int(listener.qcur or 0) == 0 and int(listener.scur or 0) == 0


class ListenerCheck(object):

    def __init__(self, app_stats, app, haproxy_count):
        self.app_stats = app_stats
        self.app = app
        self.haproxy_count = haproxy_count

    def check(self, status, predicate=lambda _: True):
        # fetch all listeners with the given status
        listeners = self.app_stats.listeners_with_status(status)
        # filter the list to contain only those listeners for which
        # `predicate(listener)` returns `True`
        listeners = [l for l in listeners if predicate(l)]
        # ensure that task and listener counts match up
        return self._check_listeners_for_tasks(listeners)

    def _check_listeners_for_tasks(self, listeners):
        """Check that the expected number of listeners are present for each task
        belonging to this application.
        """
        for task in self.app["tasks"]:
            if not self._check_listeners_for_task(listeners, task):
                return False
        return True

    def _check_listeners_for_task(self, listeners, task):
        """Check that the expected number of listeners for the given task are
        present.
        """
        # build a service name from the given marathon task which matches the
        # format used by HAProxy's stats CSV
        _prefix = task['host'].replace('.', '_')
        _port = task['ports'][0]
        svname = '{0}_{1}'.format(_prefix, _port)

        # filter listeners to only those for the given task
        listeners = [l for l in listeners if l.svname == svname]
        # ensure there's exactly one listener for each marathon-lb/haproxy
        # instance
        return len(listeners) == self.haproxy_count



# stdlib imports
import logging

# third-party imports
import click

# local imports
from shpkpr.cli import arguments, options
from shpkpr.cli.entrypoint import CONTEXT_SETTINGS
from shpkpr.template import load_values_from_environment
from shpkpr.template import render_json_template
from shpkpr.vault import resolve_secrets


logger = logging.getLogger(__name__)


@click.group('cron', context_settings=CONTEXT_SETTINGS)
def cli():
    """Manage Chronos jobs.
    """


@cli.command('show', short_help='List Chronos Jobs as json', context_settings=CONTEXT_SETTINGS)
@options.chronos_client
@options.job_name
@options.output_formatter
def show(chronos_client, job_name, output_formatter, **kw):
    """List application configuration.
    """
    jobs = chronos_client.list()

    if job_name is None:
        payload = jobs
    else:
        payload = _find_job(jobs, job_name)
    logger.info(output_formatter.format(payload))


def _inject_secrets(template, secrets):
    """Given an object containing secrets, inject them into a Chronos job prior
    to deployment.
    """
    if not template.get("environmentVariables"):
        template["environmentVariables"] = []
    for key, secret in secrets.items():
        template["environmentVariables"].append({
            "name": key,
            "value": secret,
        })
    return template


@cli.command('set', short_help='Add or Update a Chronos Job', context_settings=CONTEXT_SETTINGS)
@arguments.env_pairs
@options.chronos_client
@options.vault_client
@options.template_names
@options.template_path
@options.env_prefix
def set(chronos_client, vault_client, template_path, template_names, env_prefix, env_pairs, **kw):
    """Add or Update a job in chronos.
    """
    # use the default template if none was specified
    if not template_names:
        template_names = ["chronos/default/job.json.tmpl"]

    values = load_values_from_environment(prefix=env_prefix, overrides=env_pairs)
    current_jobs = chronos_client.list()

    for template_name in template_names:
        rendered_template = render_json_template(template_path, template_name, **values)
        resolved_secrets = resolve_secrets(vault_client, rendered_template)
        rendered_template = _inject_secrets(rendered_template, resolved_secrets)
        if _find_job(current_jobs, rendered_template['name']):
            chronos_client.update(rendered_template)
        else:
            chronos_client.add(rendered_template)


@cli.command('delete', short_help='Deletes a Job from Chronos', context_settings=CONTEXT_SETTINGS)
@arguments.job_name
@options.chronos_client
def delete(chronos_client, job_name, **kw):
    chronos_client.delete(job_name)


@cli.command('delete-tasks', short_help='Terminate all tasks for a Chronos Job.', context_settings=CONTEXT_SETTINGS)
@arguments.job_name
@options.chronos_client
def delete_tasks(chronos_client, job_name, **kw):
    chronos_client.delete_tasks(job_name)


@cli.command('run', short_help='Runs a Chronos Job', context_settings=CONTEXT_SETTINGS)
@arguments.job_name
@options.chronos_client
def run(chronos_client, job_name, **kw):
    chronos_client.run(job_name)


def _find_job(jobs, job_name):
    return list(filter(lambda j: job_name == j["name"], jobs))



# third-party imports
import click
from click.testing import CliRunner

# local imports
from shpkpr.cli import arguments


def test_env_pairs():

    @click.command()
    @arguments.env_pairs
    def greet(env_pairs):
        for k, v in env_pairs.items():
            click.echo('(%s: %s)' % (k, v))

    result = CliRunner().invoke(greet, [
        'shp=kpr',
        'MYKEY=MYVALUE',
        'MY_OTHER_KEY=MY_OTHER=VALUE',
        'EMPTY_KEY',
    ])

    assert '(shp: kpr)' in result.output
    assert '(MYKEY: MYVALUE)' in result.output
    assert '(MY_OTHER_KEY: MY_OTHER=VALUE)' in result.output
    assert '(EMPTY_KEY: )' in result.output



# stdlib imports
import datetime
import time

# local imports
from shpkpr import exceptions


class DeploymentFailed(exceptions.ShpkprException):
    pass


class DeploymentNotFound(exceptions.ShpkprException):
    pass


class MarathonDeployment(object):
    """Marathon deployment object.

    Allows the caller to check a deployment's status and cancel or rollback a
    deployment.
    """

    def __init__(self, client, deployment_id):
        self._client = client
        self.deployment_id = deployment_id

    def check(self):
        """Check if this deployment has completed.

        This method returns a True when a deployment is complete, False when a
        deployment is in progress.
        """
        try:
            self._client.get_deployment(self.deployment_id)
        except DeploymentNotFound:
            # if the deployment isn't listed, then we can consider the deployment as completed
            # successfully and return True. According to the marathon docs: "If the deployment
            # is gone from the list of deployments, then this means it is finished."
            return True

        return False

    def wait(self, timeout=900, check_interval_secs=5):
        """Waits for a deployment to finish

        If a deployment completes successfully, True is returned, if it fails
        for any reason a DeploymentFailed exception is raised. If the
        deployment does not complete within ``timeout`` seconds, a
        DeploymentFailed exception is raised.

        ``check_interval_secs`` is used to determine the delay between
        subsequent checks. The default of 5 seconds is adequate for normal
        use.
        """
        _started = datetime.datetime.utcnow()

        while True:

            time.sleep(check_interval_secs)

            # check if the deployment has completed, if it has we return True
            if self.check():
                return True

            # if the specified timeout has elapsed we raise a DeploymentFailed error
            delta = datetime.datetime.utcnow() - _started
            if delta.total_seconds() > timeout:
                raise DeploymentFailed('Timed out: %d seconds' % timeout)



# third-party imports
import requests
from cached_property import cached_property

# local imports
from . import resolver
from . import stats


class MarathonLBClient(object):
    """A high-level client used to interact with a set of load-balanced
    Marathon-LB instances.
    """

    def __init__(self, url):
        self.url = url

    @cached_property
    def urls(self):
        """Resolves the provided URL using DNS and returns a list of URLs that
        can be used to contact each individual Marathon-LB instance in a
        load-balanced set.
        """
        return resolver.resolve(self.url)

    @property
    def instance_count(self):
        """Number of Marathon-LB instances
        """
        return len(self.urls)

    def is_reloading(self):
        """Returns a boolean indicating whether any of the configured
        Marathon-LB instances is currently reloading.

        If an instance doesn't return exactly one process ID then it's either
        starting up, dead, or reloading.
        """
        for instance, pids in self._fetch_pids().items():
            if not len(pids) == 1:
                return True
        return False

    def fetch_stats(self):
        """Returns an iterable containing aggregated statistics from all
        configured Marathon-LB instances.

        See https://cbonte.github.io/haproxy-dconv/1.7/management.html#9.1 for a
        detailed breakdown of exactly what is available in the returned CSV and
        what the values in each column mean.
        """
        instance_stats = []
        for url in self.urls:
            instance_stats.append(self._fetch_instance_stats(url))
        return stats.Stats(*instance_stats)

    def _fetch_pids(self):
        """Fetch active PIDs from all Marathon-LB instances.
        """
        pids = {}
        for url in self.urls:
            pids[url] = self._fetch_instance_pids(url)
        return pids

    def _fetch_instance_pids(self, url):
        """Fetch a list of active PIDs from a single Marathon-LB instance.
        """
        response = requests.get(url + "/_haproxy_getpids")
        if not response.status_code == requests.codes.ok:
            return []
        return response.text.split()

    def _fetch_instance_stats(self, url):
        """Fetch statistics from a single Marathon-LB instance in CSV format.
        """
        response = requests.get(url + "/haproxy?stats;csv")
        response.raise_for_status()
        return response.text
